/*
 * Copyright 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <err.h>

#include "config.h"

#include <openssl/rsa.h>
#include <openssl/sha.h>
#include <openssl/dh.h>

#include <event.h>
#include <io.h>

#include "crypto.h"
#include "auth.h"
#include "dh.h"
#include "io_aux.h"

#define SERVER	1
#define CLIENT	0

#define ENCRYPT_C2S	"Encrypt|Client2Server|"
#define AUTH_C2S	"Auth|Client2Server|"
#define ENCRYPT_S2C	"Encrypt|Server2Client|"
#define AUTH_S2C	"Auth|Server2Client|"

u_char *
auth_key(struct auth_state *auth, char *string)
{
	SHA_CTX ctx;
	static u_char digest[20];

	SHA1_Init(&ctx);
	SHA1_Update(&ctx, string, strlen(string));
	SHA1_Update(&ctx, auth->shared, auth->sharedlen);
	SHA1_Update(&ctx, string, strlen(string));

	SHA1_Final(digest, &ctx);

	return (digest);
}

int
auth_compute_keys(struct auth_state *auth, int which)
{
	u_char *digest;

	BF_KEY *c2s, *s2c;
	HMAC_CTX *ac2s, *as2c;

	if (which == SERVER) {
		c2s = &auth->in;
		ac2s = &auth->authin;
		s2c = &auth->out;
		as2c = &auth->authout;
	} else {
		c2s = &auth->out;
		ac2s = &auth->authout;
		s2c = &auth->in;
		as2c = &auth->authin;
	}

	digest = auth_key(auth, ENCRYPT_C2S);
	BF_set_key(c2s, SHA_DIGEST_LENGTH, digest);
	digest = auth_key(auth, ENCRYPT_S2C);
	BF_set_key(s2c, SHA_DIGEST_LENGTH, digest);

	digest = auth_key(auth, AUTH_C2S);
	HMAC_Init(ac2s, digest, SHA_DIGEST_LENGTH, EVP_sha1());
	digest = auth_key(auth, AUTH_S2C);
	HMAC_Init(as2c, digest, SHA_DIGEST_LENGTH, EVP_sha1());

	return (0);
}

u_char *
auth_compute_hash(struct auth_state *auth, int which)
{
	struct io_buffer buf;
	SHA_CTX ctx;
	static u_char digest[SHA_DIGEST_LENGTH];

	SHA1_Init(&ctx);
	SHA1_Update(&ctx, auth->shared, auth->sharedlen);
	SHA1_Update(&ctx, auth->digest, sizeof(auth->digest));

	/* All the packets */
	memset(&buf, 0, sizeof(buf));
	if (io_buffer_bignum(&buf, auth->dh->p) != IOFILT_DONE)
		return (NULL);
	if (io_buffer_bignum(&buf, auth->dh->g) != IOFILT_DONE)
		return (NULL);
	if (which == SERVER) {
		if (io_buffer_bignum(&buf, auth->dh_pub) != IOFILT_DONE)
			return (NULL);
		if (io_buffer_bignum(&buf, auth->dh->pub_key) != IOFILT_DONE)
			return (NULL);
	} else {
		if (io_buffer_bignum(&buf, auth->dh->pub_key) != IOFILT_DONE)
			return (NULL);
		if (io_buffer_bignum(&buf, auth->dh_pub) != IOFILT_DONE)
			return (NULL);
	}
	SHA1_Update(&ctx, buf.data, buf.off);
	io_buffer_free(&buf);

	SHA1_Final(digest, &ctx);

	return (digest);
}

void
auth_callback_set(struct auth_state *auth, void (*cb)(int, short, void *))
{
	struct timeval tv;

	timeout_set(&auth->ev, cb, auth);

	timerclear(&tv);
	timeout_add(&auth->ev, &tv);
}

void
compute_dh(int fd, short which, void *arg)
{
	struct auth_state *auth = arg;

	/* We want 256 key bits out of this */
	dh_gen_key(auth->dh, 256);
}

void
auth_free_state(struct auth_state *auth)
{
	if (auth->dh != NULL)
		DH_free(auth->dh);
	if (auth->p != NULL)
		BN_free(auth->p);
	if (auth->g != NULL)
		BN_free(auth->g);
	if (auth->dh_pub != NULL)
		BN_free(auth->dh_pub);
	if (auth->shared != NULL)
		free(auth->shared);

	event_del(&auth->ev);

	free(auth);
}

int
auth_protocol_client(struct io_buffer *buf, struct auth_state *auth)
{
	u_char *sig;
	void *hash;
	size_t siglen, hashlen;
	int res;

	switch(auth->step) {
	case CAUTH_READRSAKEY:
		if (!(auth->flags & CAUTH_RRK_GOTID)) {
			res = io_buffer_readpacket(buf,
						   auth->digest,
						   sizeof(auth->digest));
			if (res != IOFILT_DONE)
				return (res);
			auth->rsa_key = rsa_find_id(auth->digest);
			if (auth->rsa_key == NULL) {
				rsa_print_id(stderr,
					     __FUNCTION__": unknown key",
					     auth->digest);
				return (IOFILT_ERROR);
			}

			auth->flags |= CAUTH_RRK_GOTID;
		}
		if (!(auth->flags & CAUTH_RRK_GOTP)) {
			res = io_buffer_getbignum(buf, &auth->p);
			if (res != IOFILT_DONE)
				return (res);
			
			auth->flags |= CAUTH_RRK_GOTP;
		}
		if (!(auth->flags & CAUTH_RRK_GOTG)) {
			res = io_buffer_getbignum(buf, &auth->g);
			if (res != IOFILT_DONE)
				return (res);
			
			auth->flags |= CAUTH_RRK_GOTG;
		}

		auth->dh = dh_new_group(auth->g, auth->p);
		auth->p = auth->g = NULL;
		dh_gen_key(auth->dh, 256);

		io_buffer_empty(buf);
		io_buffer_bignum(buf, auth->dh->pub_key);

		buf->ready = buf->off;

		auth->step = CAUTH_READDHPUB;
		auth->flags = 0;
		break;

	case CAUTH_READDHPUB:
		if (!(auth->flags & CAUTH_RDP_GOTDHPUB)) {
			res = io_buffer_getbignum(buf, &auth->dh_pub);
			if (res != IOFILT_DONE)
				return (res);
			
			auth->flags |= CAUTH_RDP_GOTDHPUB;
		}

		if (io_buffer_haspacket(buf, &hash, &hashlen) != IOFILT_DONE)
			return (IOFILT_NEEDMORE);
		auth->flags |= CAUTH_RDP_GOTSIG;

		if (hashlen != RSA_size(auth->rsa_key))
			return (IOFILT_ERROR);

		if ((sig = malloc(hashlen)) == NULL)
			return (IOFILT_ERROR);
		
		siglen = RSA_public_decrypt(hashlen, hash, sig, auth->rsa_key,
					    RSA_PKCS1_PADDING);
		if (siglen != SHA_DIGEST_LENGTH)
			goto error;

		auth->sharedlen = DH_size(auth->dh);
		auth->shared = malloc(auth->sharedlen);
		if (auth->shared == NULL)
			goto error;
		
		/* Compute the shared key */
		res = DH_compute_key(auth->shared, auth->dh_pub, auth->dh);
		if (res == -1)
			errx(1, __FUNCTION__": DH_compute_key");
		auth->sharedlen = res;

		hash = auth_compute_hash(auth, CLIENT);
		if (memcmp(hash, sig, SHA_DIGEST_LENGTH)) {
			rsa_print_id(stderr, __FUNCTION__
				     ": signature mismatch for key id",
				     auth->digest);
		error:
			free (sig);
			return (IOFILT_ERROR);
		}

		auth_compute_keys(auth, CLIENT);

		rsa_print_id(stderr, __FUNCTION__": authenticated",
			     auth->digest);
		free(sig);

		io_buffer_empty(buf);
		break;

	default:
		return (IOFILT_ERROR);
	}

	return (IOFILT_DONE);
}

int
auth_protocol_server(struct io_buffer *buf, struct auth_state *auth)
{
	u_char *hash, *sig;
	size_t sigsize, hashlen;
	int res;
	
	switch(auth->step) {
	case SAUTH_SENDRSAKEY:
		/* Empty all the data in the buffer */
		io_buffer_empty(buf);

		rsa_idkey(auth->rsa_key, auth->digest, sizeof(auth->digest));
		io_buffer_packet(buf, auth->digest, sizeof(auth->digest));

		auth->dh = choose_dh(DH_GRP_MIN, 1024, DH_GRP_MAX);
		if (auth->dh == NULL)
			return (IOFILT_ERROR);

		io_buffer_bignum(buf, auth->dh->p);
		io_buffer_bignum(buf, auth->dh->g);

		buf->ready = buf->off;

		auth->step = SAUTH_READDHPUB;
		auth_callback_set(auth, compute_dh);
		break;

	case SAUTH_READDHPUB:
		res = io_buffer_getbignum(buf, &auth->dh_pub);
		if (res != IOFILT_DONE)
			return (res);
		if (!dh_pub_is_valid(auth->dh, auth->dh_pub))
			return (IOFILT_ERROR);

		io_buffer_empty(buf);
		io_buffer_bignum(buf, auth->dh->pub_key);

		auth->sharedlen = DH_size(auth->dh);
		auth->shared = malloc(auth->sharedlen);
		if (auth->shared == NULL)
			return (IOFILT_ERROR);
		
		/* Compute the shared key */
		res = DH_compute_key(auth->shared, auth->dh_pub, auth->dh);
		if (res == -1)
			errx(1, __FUNCTION__": DH_compute_key");
		auth->sharedlen = res;

		hash = auth_compute_hash(auth, SERVER);
		if (hash == NULL)
			return (IOFILT_ERROR);

		sigsize = RSA_size(auth->rsa_key);
		if ((sig = malloc(sigsize)) == NULL)
			return (IOFILT_ERROR);

		if (RSA_private_encrypt(SHA_DIGEST_LENGTH, hash,
					sig, auth->rsa_key,
					RSA_PKCS1_PADDING) == -1)
			return (IOFILT_ERROR);

		auth_compute_keys(auth, SERVER);

		io_buffer_packet(buf, sig, sigsize);
		free(sig);

		buf->ready = buf->off;

		auth->step = SAUTH_READCLIENTID;
		break;

	case SAUTH_READCLIENTID:
		if (io_buffer_haspacket(buf, (void **)&hash,
					&hashlen) != IOFILT_DONE)
			return (IOFILT_NEEDMORE);
		if (hashlen != sizeof(auth->client_id))
			return (IOFILT_ERROR);
		memcpy(auth->client_id, hash, hashlen);

		/* We might get more data from the client batched into this
		 * response.
		 */
		io_buffer_drainpacket(buf);
		
		auth->step = -1;
		break;

	default:
		return (IOFILT_ERROR);
	}

	return (IOFILT_DONE);
}
