/*
 * Copyright 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <err.h>

#include <openssl/rsa.h>
#include <openssl/dh.h>
#include <openssl/sha.h>

#include "config.h"

#include <event.h>
#include <io.h>

#include "io_aux.h"
#include "auth.h"
#include "crypto.h"
#include "client.h"
#include "job.h"

/* Prototypes */
int client_authenticate(struct io_duplex *);

char *progname;

/* Plaintext data handler */
struct io_filter *src = NULL;
struct io_filter *dst = NULL;
struct io_filter *cmdhandler = NULL;

struct io_obj *stddst;
struct io_filter *anchor;

/* Permanent client id */
u_char ourid[16];
int ourid_ready = 0;

/* Pre-configured server */
char *mastername = "127.0.0.1";
u_short masterport = 8010;

struct event evcheck;

int authenticated = 0;

#define MSTCHECK_INTERVAL	30

void
master_checkadd(void)
{
	struct timeval tv;
	u_int32_t tmp = arc4random();

	timerclear(&tv);
	tv.tv_sec = MSTCHECK_INTERVAL + (tmp % (MSTCHECK_INTERVAL/3));
	timeout_add(&evcheck, &tv);
}

void
master_check(int fd, short which, void *arg)
{
	struct io_duplex *dplx;

	if (authenticated)
		goto out;

	dplx = io_new_connector(mastername, masterport);
	if (dplx == NULL) {
		warnx(__FUNCTION__": io_new_connection(%s, %d)",
		      mastername, masterport);
		goto out;
	}

	if (client_authenticate(dplx) == -1) {
		io_terminate(dplx->io_src, IORES_ERROR);
		goto out;
	}

	fprintf(stderr, __FUNCTION__": authenticating to %s:%d\n",
		mastername, masterport);
	return;

 out:
	master_checkadd();
}

void
master_dead(struct io_obj *obj, void *arg)
{
	authenticated = 0;

	fprintf(stderr, __FUNCTION__": server is gone\n");

	auth_free_state(arg);

	src = dst = NULL;

	client_masterdead();
	master_checkadd();
}

int
filter_authprotocol(void *arg, struct io_filter *filter,
		    struct io_buffer *buf, int flags)
{
	struct auth_state *auth = arg;
	struct io_buffer *mybuf = &filter->io_buf;
	int res;

	if (io_buffer_append(mybuf, buf) == -1)
		return (-1);

	res = auth_protocol_client(mybuf, auth);

	if (res == IOFILT_DONE && mybuf->off == 0) {
		struct io_filter *filt;

		authenticated = 1;

		/* Detach the authenticator */
		io_detach(auth->dplx->io_src, filter);
		io_detach(filter, auth->dplx->io_dst);

		filt = io_new_filter(filter_encrypt, auth);
		io_attach(filt, auth->dplx->io_dst);
		dst = filt;

		filt = io_new_filter(filter_decrypt, auth);
		io_attach(auth->dplx->io_src, filt);
		src = filt;

		cmdhandler = io_new_filter(filter_commandhandler, NULL);
		io_attach(src, cmdhandler);

		/* Send permanent id to finish protocol */
		if (ourid_ready) {
			int i;
			SHA_CTX ctx;
			u_char digest[20];

			/* For operating systems that suck we need this */
			SHA1_Init(&ctx);
			SHA1_Update(&ctx, auth, sizeof(*auth));
			SHA1_Final(digest, &ctx);

			for (i = 0; i < sizeof(ourid); i++)
				ourid[i] ^= digest[i];
			
			ourid_ready = 1;
		}
		io_filter_packet(dst, ourid, sizeof(ourid));

		/* Send all data that we have accumulated */
		client_bufferflush((struct io_obj *)dst);

		client_sendstatus();

		/* Causes this filter to be cleaned up */
		return (IOFILT_END);
	}

	return (res);
}

int
client_authenticate(struct io_duplex *dplx)
{
	struct io_filter *filt;
	struct auth_state *auth;

	auth = calloc(1, sizeof (struct auth_state));
	if (auth == NULL) {
		warn(__FUNCTION__": calloc");
		return (-1);
	}
	auth->dplx = dplx;

	filt = io_new_filter(filter_authprotocol, auth);

	dplx->io_src->io_free = master_dead;
	dplx->io_src->io_freearg = auth;

	/* This filter reads from the source, and writes back */
	io_attach(dplx->io_src, filt);
	io_attach(filt, dplx->io_dst);

	return (0);
}

int
filter_handlenew(void *arg, struct io_filter *filter, struct io_buffer *buf, 
		 int flags)
{
	struct io_buffer *mybuf = &filter->io_buf;
	struct io_duplex *dplx;
	int newfd;

	if (buf->ready < sizeof (int))
		return (IOFILT_NEEDMORE);

	if (io_buffer_extend(mybuf, 256) == -1)
		return (IOFILT_ERROR);

	newfd = *(int *)buf->data;

	if ((dplx = io_new_socket(newfd)) == NULL)
		return (IOFILT_ERROR);

	if (authenticated) {
		char *message = "A master controller is already attached.\n"
			"Go away, please.\n";
		io_data_write(dplx->io_dst, message, strlen(message));
		io_obj_close(dplx->io_dst);

		snprintf(mybuf->data, 256, "Connection on fd %d denied\n",
			 newfd);
	} else {
		if (client_authenticate(dplx) == -1) {
			io_terminate(dplx->io_src, IORES_ERROR);
			return (IOFILT_NEEDMORE);
		}
	
		snprintf(mybuf->data, 256, "New connection on fd %d\n", newfd);
	}

	mybuf->ready = mybuf->off = strlen(mybuf->data);

	return (IOFILT_DONE);
}

void
create_id(u_int32_t *buf)
{
	int i;

	for (i = 0; i < 4; i++)
		buf[i] = arc4random();
}

void
usage(void)
{
	fprintf(stderr, "%s: [-s server] [-c cadir] [-l logfile]\n",
	    progname);
	exit(1);
}

void
client_closefds(void)
{
	int i;

	for (i = 3; i < 255; i++)
		close(i);
}

int
main(int argc, char **argv)
{
	extern char *optarg;
	extern int optind;
	struct io_obj *fdsrc;
	struct io_filter *filt;
	char *cadirname = "ca", *p;
	char *logfile = "client.stderr";
	struct timeval tv;
	int ch;

	/* Ugly hack to allow remote update - XXX temporary */
	client_closefds();

	progname = argv[0];

	while ((ch = getopt(argc, argv, "c:s:l:")) != -1) {
		switch (ch) {
		case 's':
			p = strchr(optarg, ':');
			if (p == NULL) {
				mastername = optarg;
				break;
			}

			*p = '\0';
			mastername = optarg;
			masterport = atoi(p+1);

			if (masterport == 0)
				usage();

			break;

		case 'c':
			cadirname = optarg;
			break;
			    
		case 'l':
			logfile = optarg;
			break;

		default:
			usage();
		}
	}

	argc -= optind;
	argv += optind;

	crypto_addrandom();
	
	/* Create our permanent identifier */
	create_id((u_int32_t *)ourid);

	event_init();
	
	job_init();

	/* read server keys */
	rsa_list_init();
	rsa_read_dir(cadirname);

	anchor = io_new_filter(NULL, NULL);

	/* stderr is gone, reopen it */
	if (freopen(logfile, "a", stderr) == NULL)
		err(1, "freopen(%s)", logfile);
	else {
		time_t now = time(NULL);
		setvbuf(stderr, NULL, _IOLBF, 0);
		fprintf(stderr, "Restart at %s", ctime(&now));
		fflush(stderr);
	}

	stddst = io_new_sink(fileno(stderr));

	/* Protect stdin and stdout */
	io_attach(anchor, stddst);

	fdsrc = io_new_listener("0.0.0.0", 8080);
	if (fdsrc != NULL) {
		filt = io_new_filter(filter_handlenew, NULL);
		io_attach(fdsrc, filt);
		io_attach(filt, stddst);
	}

	timeout_set(&evcheck, master_check, NULL);
	timerclear(&tv);
	timeout_add(&evcheck, &tv);

	event_dispatch();

	exit(0);
}
