#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>

#include <openssl/rsa.h>
#include <openssl/dh.h>

#include "config.h"

#include <event.h>
#include <io.h>

#include "crypto.h"
#include "auth.h"
#include "client.h"
#include "io_aux.h"
#include "commands.h"
#include "util.h"
#include "job.h"

#define MAXNAME	1024

/* This is how we communicate */
extern struct io_filter *src;
extern struct io_filter *dst;

/* Temporary buffer, we use when the server is gone */
static struct io_buffer databuf;

/* Handle for the command handler */
extern struct io_filter *cmdhandler;

int cmdexecuting;

/* Data used for executing the current request */
struct io_filter *cmd_filt = NULL;
char cmd_jobid[MAXNAME];
char cmd_file[MAXNAME];
int cmd_fd = -1;
u_int64_t cmd_sz;
u_int64_t cmd_count;
void (*cmd_abortcb)(void) = NULL;

/* Responses from the client on commands need to be 0-terminated */
struct command {
	char *cmd;
	int (*func)(char *, size_t, char *);
};

int client_command_init(char *, size_t, char *);
int client_command_run(char *, size_t, char *);
int client_command_query(char *, size_t, char *);

struct command commands[] = {
	{
		CMD_QUERY,
		client_command_query
	},
	{
		CMD_INIT,
		client_command_init
	},
	{
		CMD_RUN,
		client_command_run
	},
	{
		NULL, NULL
	}
};

char tmpbuf[MAXNAME];
char cmdbuf[MAXNAME];

int
client_command_query(char *buf, size_t size, char *cmd)
{
	char hash[41], filename[256];
	struct dataset *ds;
	struct job *job;
	char *jobid, *file, *arg;
	struct stat sb;

	cmd += CMD_SZ(CMD_QUERY);
	cmd += strspn(cmd, " ");

	strlcpy(tmpbuf, cmd, sizeof(tmpbuf));

	arg = tmpbuf;
	jobid = strnsep(&arg, " ");

	if (arg == NULL || strlen(arg) == 0)
		goto error;

	if (stat(jobid, &sb) == -1 || !S_ISDIR(sb.st_mode)) {
		snprintf(buf, size, "%s is not a directory", jobid);
		return (IOFILT_ERROR);
	}

	if ((job = job_find(jobid)) == NULL) {
		snprintf(buf, size, "job_find(%s) failed", jobid);
		return (IOFILT_ERROR);
	}

	file = strnsep(&arg, " ");
	if (file == NULL || strlen(file) == 0) {
	error:
		snprintf(buf, size, "bad argument");
		return (IOFILT_ERROR);
	}

	/* Add file to job */
	strlcpy(filename, jobid, sizeof(filename));
	strlcat(filename, "/", sizeof(filename));
	strlcat(filename, file, sizeof(filename));

	ds = dataset_findfile(job->job_files, &job->job_nfiles,
			      NULL, filename);
	if (ds == NULL) {
		snprintf(buf, size, "%s unknown", filename);
		return (IOFILT_ERROR);
	}

	bin2hex(hash, ds->sf_hash, sizeof(ds->sf_hash));
	snprintf(buf, size, "%s %lld %s",
		 ds->sf_name, ds->sf_size, hash);

	client_sendstatus();
	return (IOFILT_DONE);
}

int
client_command_run(char *buf, size_t size, char *cmd)
{
	struct job *job;
	char *jobid, *arg;
	struct stat sb;
	char *argv[4];

	cmd += CMD_SZ(CMD_RUN);
	cmd += strspn(cmd, " ");

	strlcpy(tmpbuf, cmd, sizeof(tmpbuf));

	arg = tmpbuf;
	jobid = strnsep(&arg, " ");

	if (!file_checkname(jobid)) {
		snprintf(buf, size, "%s is not a safe filename", jobid);
		return (IOFILT_ERROR);
	}
	
	if (stat(jobid, &sb) == -1 || !S_ISDIR(sb.st_mode)) {
		snprintf(buf, size, "%s is not a directory", jobid);
		return (IOFILT_ERROR);
	}

	if ((job = job_find(jobid)) == NULL &&
	    (job = job_new(jobid)) == NULL) {
		snprintf(buf, size, "job_new(%s) failed", jobid);
		return (IOFILT_ERROR);
	}

	if (job->job_pid != 0) {
		snprintf(buf, size, "job %s already running at %d",
			 jobid, job->job_pid);
		return (IOFILT_ERROR);
	}

	/* Execute command via shell, that way we can get expansion, etc... */
	argv[0] = _PATH_SH;
	argv[1] = "-c";
	argv[2] = arg;
	argv[3] = NULL;

	if (job_fork_cmd(job, argv[0], argv) == -1) {
		snprintf(buf, size, "job_fork_cmd(%s) failed", jobid);
		return (IOFILT_ERROR);
	}

	client_sendstatus();
	return (IOFILT_DONE);
}

int
client_untar(char *jobid, char *file)
{
	char *argv[4];
#ifdef HAVE_BROKEN_TAR
	char tmp[256];
#endif

	if (!file_hasextension(file, ".tar.gz") &&
	    !file_hasextension(file, ".tgz"))
		return (0);

#ifdef HAVE_BROKEN_TAR
	/* Fucked up operating systems deserve to die */
	snprintf(tmp, sizeof(tmp), "%s -dc %s | %s -xf -",
		 _PATH_GZIP, file, _PATH_TAR);
	argv[0] = _PATH_SH;
	argv[1] = "-c";
	argv[2] = tmp;
	argv[3] = NULL;
	return (job_execute_cmd(jobid, _PATH_SH, argv));
#else
	argv[0] = _PATH_TAR;
	argv[1] = "-xzf";
	argv[2] = file;
	argv[3] = NULL;
	return (job_execute_cmd(jobid, _PATH_TAR, argv));
#endif
}

void
client_init_abort(void)
{
	close(cmd_fd);
	io_detach(cmdhandler, cmd_filt);
	io_obj_close((struct io_obj *)cmd_filt);
}

int
filter_clientinit(void *arg, struct io_filter *filter,
		  struct io_buffer *buf, int flags)
{
	struct job *job;
	ssize_t n;

	n = write(cmd_fd, buf->data, buf->ready);
	if (n == -1 || n != buf->ready) {
		warn(__FUNCTION__": write: %d", n);
		return (IOFILT_ERROR);
	}

	cmd_count += buf->ready;

	/* Are we done with the data transfer? */
	if (cmd_count == cmd_sz) {
		close(cmd_fd);
		cmd_fd = -1;
		cmd_abortcb = NULL;

		/* Detach from command handler */
		io_detach(cmdhandler, filter);
		cmdexecuting = 0;

		fprintf(stderr, __FUNCTION__": transfer done: %lld bytes\n",
			cmd_count);

		if (client_untar(cmd_jobid, cmd_file) == -1) {
			snprintf(tmpbuf, sizeof(tmpbuf),
				 "%s: client_untar %s",
				 CMD_ERROR, cmd_file);
			io_filter_packet(dst, tmpbuf, strlen(tmpbuf) + 1);
			return (IOFILT_END);
		}

		if ((job = job_find(cmd_jobid)) == NULL &&
		    (job = job_new(cmd_jobid)) == NULL) {
			snprintf(tmpbuf, sizeof(tmpbuf),
				 "%s: job_new(%s) failed",
				 CMD_ERROR, cmd_jobid);
			io_filter_packet(dst, tmpbuf, strlen(tmpbuf) + 1);
			return (IOFILT_END);
		}

		/* Add file to job */
		strlcpy(tmpbuf, cmd_jobid, sizeof(tmpbuf));
		strlcat(tmpbuf, "/", sizeof(tmpbuf));
		strlcat(tmpbuf, cmd_file, sizeof(tmpbuf));

		/* Keep track of configured files */
		if (dataset_addfile(job->job_files, &job->job_nfiles,
				    tmpbuf, NULL) == NULL) {
			snprintf(tmpbuf, sizeof(tmpbuf),
				 "%s: client_untar %s",
				 CMD_ERROR, cmd_file);
			io_filter_packet(dst, tmpbuf, strlen(tmpbuf) + 1);
			return (IOFILT_END);
		}

		io_filter_packet(dst, CMD_DONE, sizeof(CMD_DONE));

		client_sendstatus();

		/* Remove us */
		return (IOFILT_END);
	}

	return (IOFILT_NEEDMORE);
}

/* Send the data to the server directly, or buffer it */

void
client_bufferorsend(char *data, size_t size)
{
	if (dst != NULL)
		io_filter_packet(dst, data, size);
	else
		io_buffer_packet(&databuf, data, size);
}

void
client_bufferflush(struct io_obj *dst)
{
	if (databuf.off == 0)
		return;

	databuf.ready = databuf.off;
	io_buffer_write(dst, &databuf);

	io_buffer_drain(&databuf, databuf.ready);
}

void
client_jobresult(struct job *job, int died)
{
	snprintf(tmpbuf, sizeof(tmpbuf),
		 "%s %d %s",
		 CMD_INFO,
		 died ? INFO_CHLDDEAD : INFO_CHLDDONE,
		 job->job_id);

	client_bufferorsend(tmpbuf, strlen(tmpbuf) + 1);

	client_sendstatus();
}

int
client_command_init(char *buf, size_t size, char *cmd)
{
	mode_t mode, dir_mode;
	struct stat sb;

	cmd += CMD_SZ(CMD_INIT);
	cmd += strspn(cmd, " ");

	cmd_jobid[MAXNAME-1] = cmd_file[MAXNAME-1] = '\0';
	if (sscanf(cmd, "%1023s %1023s %lld",
		   cmd_jobid, cmd_file, &cmd_sz) != 3) {
		snprintf(buf, size, "command badly formatted");
		return (IOFILT_ERROR);
	}

	if (!file_checkname(cmd_jobid)) {
		snprintf(buf, size, "bad jobid");
		return (IOFILT_ERROR);
	}

	if (!file_checkname(cmd_file)) {
		snprintf(buf, size, "bad file name");
		return (IOFILT_ERROR);
	}
	
	mode = 0777 & ~umask(0);
	dir_mode = mode | S_IWUSR | S_IXUSR;

	if (stat(cmd_jobid, &sb) != -1) {
		/* File exists */
		if (!S_ISDIR(sb.st_mode)) {
			snprintf(buf, size,
				 "%s is not a directory", cmd_jobid);
			return (IOFILT_ERROR);
		}
	} else {
		if (errno != ENOENT) {
			snprintf(buf, size, "%s", strerror(errno));
			return (IOFILT_ERROR);
		}

		if (mkdir(cmd_jobid, dir_mode) == -1) {
			snprintf(buf, size, "mkdir: %s", strerror(errno));
			return (IOFILT_ERROR);
		}
	}

	strlcpy(tmpbuf, cmd_jobid, sizeof(tmpbuf));
	strlcat(tmpbuf, "/", sizeof(tmpbuf));
	strlcat(tmpbuf, cmd_file, sizeof(tmpbuf));

	cmd_fd = open(tmpbuf, O_CREAT|O_TRUNC|O_WRONLY, mode);
	if (cmd_fd == -1) {
		snprintf(buf, size, "open: %s", strerror(errno));
		return (IOFILT_ERROR);
	}
	cmd_count = 0;

	cmd_filt = io_new_filter(filter_clientinit, NULL);
	if (cmd_filt == NULL)
		errx(1, __FUNCTION__": io_new_filter");

	io_attach(cmdhandler, cmd_filt);

	return (IOFILT_NEEDMORE);
}

void
client_sendstatus(void)
{
	char *jobs;

	jobs = job_statusline();

	snprintf(tmpbuf, sizeof(tmpbuf), "%s: %s", CMD_STATUS, jobs);

	client_bufferorsend(tmpbuf, strlen(tmpbuf) + 1);
}

void
client_masterdead(void)
{
	if (cmdexecuting == 0)
		return;

	fprintf(stderr, __FUNCTION__": aborting command\n");
	if (cmd_abortcb != NULL)
		(*cmd_abortcb)();

	cmdexecuting = 0;
}

int
filter_commandhandler(void *arg, struct io_filter *filter,
		      struct io_buffer *buf, int flags)
{
	int res;
	void *data;
	size_t size;
	struct command *cmd;

	res = io_buffer_haspacket(buf, &data, &size);
	if (cmdexecuting) {
		struct io_buffer *mybuf = &filter->io_buf;

		if (res == IOFILT_DONE && size == sizeof(CMD_ABORT) &&
		    strcasecmp(data, CMD_ABORT) == 0) {
			if (cmd_abortcb != NULL)
				cmd_abortcb();

			cmdexecuting = 0;
			return (IOFILT_NEEDMORE);
		}

		if (io_buffer_append(mybuf, buf) == IOFILT_ERROR)
			return (IOFILT_ERROR);

		mybuf->ready = mybuf->off;

		return (IOFILT_DONE);
	}

	if (res != IOFILT_DONE)
		return (res);

	if (((char *)data)[size - 1] != '\0') {
		fprintf(stderr, __FUNCTION__": illegal command from server\n");
		return (IOFILT_ERROR);
	}

	fprintf(stderr, "Received: command %s\n", (char *)data);

	for(cmd = commands; cmd->cmd; cmd++) {
		/* Find out what command was sent. */
		if (strncasecmp(cmd->cmd, data, strlen(cmd->cmd)) == 0)
			break;
	}

	/* Unknown command */
	if (cmd->cmd == NULL) {
		snprintf(tmpbuf, sizeof(tmpbuf),
			 "%s: Unknown command: %s", CMD_ERROR, (char *)data);
		goto done;
	}

	cmdbuf[0] = '\0';
	res = cmd->func(cmdbuf, sizeof(cmdbuf), data);

	/* Preparation failed */
	if (res == IOFILT_ERROR) {
		snprintf(tmpbuf, sizeof(tmpbuf),
			 "%s: failure on %s: %s", CMD_ERROR, (char *)data,
			 cmdbuf);
		goto done;
	}

	if (res == IOFILT_NEEDMORE)
		cmdexecuting = 1;

	snprintf(tmpbuf, sizeof(tmpbuf), "%s: %s",
		 res == IOFILT_NEEDMORE ? CMD_OK : CMD_DONE, cmdbuf);

 done:
	io_filter_packet(dst, tmpbuf, strlen(tmpbuf) + 1);
	return (IOFILT_NEEDMORE);
}
