/*
 * Copyright 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>

#include "config.h"

#include <event.h>
#include <io.h>

#include "commands.h"
#include "server.h"
#include "ui.h"
#include "io_aux.h"

char tmpbuf[256];

/*
 * Creates a new serialization queue for commands.
 */

struct client_serialize_queue *
command_serialize_new(int max)
{
	struct client_serialize_queue *csq;

	if ((csq = calloc(1, sizeof(struct client_serialize_queue))) == NULL)
		err(1, __FUNCTION__": calloc");
	TAILQ_INIT(&csq->csq_queue);

	csq->csq_max = max;

	return (csq);
}

/* 
 * Put the last command in the client's command queue on a serialized
 * command queue.  A serialized command queue allows only a certain
 * number of the same command to run at the same time, e.g., file
 * transfers.
 */

void
command_serialize_last(struct client_serialize_queue *csq,
    struct client *client)
{
	struct cmdlist *head = &client->client_cmdqueue;
	struct client_cmd *cmd;

	if ((cmd = TAILQ_LAST(head, csq)) == NULL)
		err(1, __FUNCTION__": no command on queue");
	if (cmd->csq_parent != NULL)
		err(1, __FUNCTION__": command already serialized");

	TAILQ_INSERT_TAIL(&csq->csq_queue, cmd, cc_csq_next);
	cmd->csq_parent = csq;
	csq->csq_pending++;
}

int
command_serialize_start(struct client_cmd *cmd)
{
	struct client_serialize_queue *csq = cmd->csq_parent;

	if (csq->csq_current >= csq->csq_max)
		return (-1);

	cmd->cc_flags |= CCFLAG_STARTED;
	TAILQ_REMOVE(&csq->csq_queue, cmd, cc_csq_next);

	csq->csq_current++;
	csq->csq_pending--;

	return (0);
}

void
command_serialize_remove(struct client_cmd *cmd)
{
	struct client_serialize_queue *csq = cmd->csq_parent;
	struct client_cmd *next;

	if (!(cmd->cc_flags & CCFLAG_STARTED)) {
		TAILQ_REMOVE(&csq->csq_queue, cmd, cc_csq_next);
		csq->csq_pending--;
		return;
	}

	csq->csq_current--;

	while (csq->csq_current < csq->csq_max) {
		for (cmd = TAILQ_FIRST(&csq->csq_queue); cmd; cmd = next) {
			next = TAILQ_NEXT(cmd, cc_csq_next);

			if (command_start(cmd->cc_client) != -1)
				break;
		}

		if (cmd == NULL)
			break;
	}
}

struct client_cmd *
command_script_start(struct cmdlist *head)
{
	return (TAILQ_LAST(head, cmdlist));
}

void
command_script_end(struct cmdlist *head, struct client_cmd *start)
{
	struct client_cmd *end = TAILQ_LAST(head, cmdlist);
	
	if (end == NULL || end == start) {
		fprintf(stderr, __FUNCTION__": no commands in script\n");
		return;
	}

	if (start == NULL)
		start = TAILQ_FIRST(head);
	else
		start = TAILQ_NEXT(start, cc_next);

	while (start != end) {
		start->cc_flags |= CCFLAG_SCRIPT;

		start = TAILQ_NEXT(start, cc_next);
	}

	end->cc_flags |= CCFLAG_SCRIPT | CCFLAG_SCRIPTEND;
}

/*
 * Removes pending commands after current command if a command in a
 * script failed to execute properly.
 */

void
command_script_remove(struct client *client, struct client_cmd *start)
{
	struct cmdlist *head = &client->client_cmdqueue;
	struct client_cmd *next;
	int end = 0;

	if (start == NULL)
		start = TAILQ_FIRST(head);
	else if (start->cc_flags & CCFLAG_SCRIPTEND)
		return;
	else
		start = TAILQ_NEXT(start, cc_next);

	for (; start != NULL && !end; start = next) {
		next = TAILQ_NEXT(start, cc_next);

		if (start->cc_flags & CCFLAG_SCRIPTEND)
			end = 1;

		command_free(client, start);
	}
}

int
command_queue(struct cmdlist *head, char *command,
	      int (*cb)(int, struct client *, void *, char*), void *arg)
{
	struct client_cmd *cc;

	cc = calloc(1, sizeof(struct client_cmd));
	if (cc == NULL)
		return (-1);

	strlcpy(cc->cc_command, command, sizeof(cc->cc_command));
	cc->cc_cb = cb;
	cc->cc_cbarg = arg;

	TAILQ_INSERT_TAIL(head, cc, cc_next);

	return (0);
}

void
command_dequeue(struct cmdlist *head, struct client_cmd *cc)
{
	TAILQ_REMOVE(head, cc, cc_next);

	if (cc->csq_parent != NULL)
		command_serialize_remove(cc);

	free(cc);
}

void
command_free(struct client *client, struct client_cmd *cc)
{
	cc->cc_cb(-1, client, cc->cc_cbarg, NULL);

	command_dequeue(&client->client_cmdqueue, cc);
}

int
command_start(struct client *client)
{
	struct cmdlist *head = &client->client_cmdqueue;
	struct client_cmd *cc;

	if (client->flags & CLIENT_CMDINPROGRESS)
		return (-1);

	if ((cc = TAILQ_FIRST(head)) == NULL)
		return (-1);
	
	cc->cc_client = client;

	if (cc->csq_parent != NULL && command_serialize_start(cc) == -1)
		return (-1);

	client->flags |= CLIENT_CMDINPROGRESS;

	io_filter_packet(client->dst,
			 cc->cc_command, strlen(cc->cc_command) + 1);

	client_settimeout(client);
	return (0);
}

int
command_reply(struct client *client, char *line, int res)
{
	struct client_cmd *cc;

	if (!(client->flags & CLIENT_CMDINPROGRESS)) {
		snprintf(tmpbuf, sizeof(tmpbuf), __FUNCTION__
			 ": got cmd response but not waiting: %s:%u\n",
			 client->name, client->port);
		ui_infocb(client, tmpbuf);
		return (IOFILT_ERROR);
	}

	cc = TAILQ_FIRST(&client->client_cmdqueue);
	if (cc == NULL)
		return (IOFILT_ERROR);

	/* Remove timeout */
	timeout_del(&client->ev);

	/* The callback will tell us, if we need to wait for more */
	res = (cc->cc_cb)(res, client, cc->cc_cbarg, line);
	if (res == IOFILT_NEEDMORE)
		return (IOFILT_NEEDMORE);

	if (res == IOFILT_ERROR && (cc->cc_flags & CCFLAG_SCRIPT))
		command_script_remove(client, cc);

	client->flags &= ~(CLIENT_CMDINPROGRESS|CLIENT_CMDREADY);
	command_dequeue(&client->client_cmdqueue, cc);
	return (IOFILT_DONE);
}
