/*
 * Copyright 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <dirent.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
#include <fcntl.h>

#include <netinet/in.h>

#include <openssl/pem.h>
#include <openssl/rsa.h>
#include <openssl/rand.h>

#include <event.h>
#include <io.h>

#include "config.h"
#include "crypto.h"
#include "util.h"
#include "auth.h"
#include "io_aux.h"

#ifndef USE_LOG
#define LOG_DBG(x)
#endif

#ifndef roundup
#define        roundup(x, y)   ((((x)+((y)-1))/(y))*(y))
#endif
#ifndef rounddown
#define        rounddown(x, y)   (((x)/(y))*(y))
#endif

#ifndef PATH_MAX
#define PATH_MAX	1024
#endif /* !PATH_MAX */

TAILQ_HEAD(rsa_queue, rsa_keyid) rsalist;

void
rsa_list_init()
{
	TAILQ_INIT(&rsalist);
}

int
rsa_read_dir(char *name)
{
	DIR *dir;
	struct dirent *file;
	char fullname[PATH_MAX];
	int off;

	if (strlen(name) >= sizeof (fullname) - 2)
		errx(1, __FUNCTION__": directory name too long");

	dir = opendir(name);
	if (dir == NULL)
		err(1, "opendir");

	strlcpy(fullname, name, sizeof (fullname));
	off = strlen(fullname);
	if (fullname[off - 1] != '/') {
		strlcat(fullname, "/", sizeof(fullname));
		off++;
	}

	while ((file = readdir(dir)) != NULL) {
		RSA *tmpkey;
		struct rsa_keyid *tmpid;

		strlcpy(fullname + off, file->d_name, sizeof(fullname) - off);

#ifdef HAVE_GOOD_DIRENT
		/* NFS does not have a d_type */
		if (file->d_type != DT_UNKNOWN) {
			if (file->d_type != DT_REG && file->d_type != DT_LNK)
				continue;
		} else {
#else
		{
#endif
			struct stat sb;

			if (stat(fullname, &sb) == -1 ||
			    !(sb.st_mode & S_IFREG))
				continue;
		}

		tmpkey = rsa_read_key(fullname, RSA_PUBLIC);
		if (tmpkey == NULL)
			continue;

		tmpid = malloc(sizeof(struct rsa_keyid));
		if (tmpid == NULL)
			err(1, "malloc");

		tmpid->key = tmpkey;
		rsa_idkey(tmpkey, tmpid->digest, sizeof(tmpid->digest));

		TAILQ_INSERT_TAIL(&rsalist, tmpid, rsa_next);
	}

	return (TAILQ_FIRST(&rsalist) != NULL ? 0 : -1);
}

RSA *
rsa_find_id(u_char *digest)
{
	struct rsa_keyid *keyid;

	TAILQ_FOREACH(keyid, &rsalist, rsa_next) {
		if (!memcmp(keyid->digest, digest, sizeof(keyid->digest)))
			break;
	}

	if (keyid == NULL)
		return (NULL);

	return (keyid->key);
}

void
rsa_print_id(FILE *fp, char *text, u_char *digest)
{
	char hexdigest[41];

	hexdigest[40] = '\0';

	bin2hex(hexdigest, digest, 20);
	fprintf(fp, "%s: 0x%s\n", text, hexdigest);
}

RSA *
rsa_read_key(char *name, enum keytype type)
{
	FILE *fp;
	RSA *key;

	if ((fp = fopen(name, "r")) == NULL)
		return (NULL);

	LOG_DBG((LOG_CRYPTO, 20, __FUNCTION__": reading key from %s", name));

	if (type == RSA_PRIVATE)
		key = PEM_read_RSAPrivateKey(fp, NULL, NULL, NULL);
	else
		key = PEM_read_RSAPublicKey(fp, NULL, NULL, NULL);

	if (key == NULL)
		warnx(__FUNCTION__": key in %s corrupt", name);

	fclose(fp);
	return (key);
}

void
rsa_write_key(char *name, RSA *key, enum keytype type)
{
	FILE *fp;
	int fd;

	LOG_DBG((LOG_CRYPTO, 20, __FUNCTION__": writing key to %s", name));

	fd = open(name, O_WRONLY | O_CREAT | O_TRUNC,
		  type == RSA_PRIVATE ? 0600: 0644);
	if (fd < 0)
		err(1, "open %s failed", name);
		
	fp = fdopen(fd, "w");
	if (fp == NULL )
		err(1, "fdopen");

	if (type == RSA_PRIVATE) {
		if (PEM_write_RSAPrivateKey(fp, key, NULL, NULL, 0,
					    NULL, NULL) == 0)
			errx(1, "PEM_write_RSAPrivateKey");
	} else {
		if (PEM_write_RSAPublicKey(fp, key) == 0)
			errx(1, "PEM_write_RSAPublicKey");
	}
	fclose(fp);
}

int
rsa_idkey(RSA *key, u_char *data, size_t size)
{
	SHA_CTX ctx;
	int len;
	u_char *tmp;
	u_char digest[20];

	len = BN_num_bytes(key->n);
	if ((tmp = malloc(len)) == NULL) {
		warn(__FUNCTION__": malloc");
		return (-1);
	}

	if (BN_bn2bin(key->n, tmp) != len) {
		free(tmp);
		return (-1);
	}

	/* Create SHA Id */
	SHA1_Init(&ctx);
	SHA1_Update(&ctx, tmp, len);
	SHA1_Final(digest, &ctx);
	free(tmp);

	if (sizeof(digest) < size)
		size = sizeof(digest);
	memcpy(data, digest, size);

	return (size);
}

int
filter_encrypt(void *arg, struct io_filter *filter,
	       struct io_buffer *buf, int flags)
{
	struct io_buffer *mybuf = &filter->io_buf;
	struct auth_state *auth = arg;
	HMAC_CTX ctx;
	u_char pad[8], *indata, *outdata;

	int padlen, totlen, rlen, i;
	u_int32_t len;

	if (buf->ready == 0)
		return (IOFILT_NEEDMORE);

	padlen = 8 - (buf->ready % 8);
	totlen = sizeof(u_int32_t) + buf->ready + padlen + SHA_DIGEST_LENGTH;

	if (io_buffer_extend(mybuf, totlen) == -1)
		return (IOFILT_ERROR);

	outdata = mybuf->data + mybuf->off + sizeof(u_int32_t);
	indata= buf->data;

	rlen = rounddown(buf->ready, 8);
	if (rlen) {
		BF_cbc_encrypt(indata, outdata, rlen, &auth->out,
			       auth->outiv, 1);

		memcpy(auth->outiv, outdata + rlen - 8, 8);

		indata += rlen;
		outdata += rlen;
	}

	/* Do the rest */
	rlen = buf->ready % 8;
	for (i = 0; i < padlen; i++) {
		pad[rlen + i] = i;
	}
	memcpy(pad, indata, rlen);
	BF_cbc_encrypt(pad, outdata, 8, &auth->out, auth->outiv, 1);
	memcpy(auth->outiv, outdata, 8);

	outdata += 8;

	/* Write the length */
	len = htonl(totlen - sizeof(u_int32_t));
	memcpy(mybuf->data + mybuf->off, &len, sizeof(u_int32_t));

	ctx = auth->authout;
	HMAC_Update(&ctx, mybuf->data + mybuf->off,
		    totlen - SHA_DIGEST_LENGTH);

	len = SHA_DIGEST_LENGTH;
	HMAC_Final(&ctx, outdata, &len);

	mybuf->off += totlen;
	mybuf->ready = mybuf->off;

	return (IOFILT_DONE);
}

int
filter_decrypt(void *arg, struct io_filter *filter,
	       struct io_buffer *buf, int flags)
{
	struct io_buffer *mybuf = &filter->io_buf;
	struct io_buffer *tmpbuf = &filter->io_tmpbuf;
	struct auth_state *auth = arg;
	u_char *data;
	size_t len, dlen;
	u_char digest[SHA_DIGEST_LENGTH], iv[8];
	HMAC_CTX ctx;
	int i;

	if (io_buffer_append(tmpbuf, buf) == -1)
		return (IOFILT_ERROR);

	mybuf->off = 0;

	for (;
	     io_buffer_haspacket(tmpbuf, (void **)&data, &len) == IOFILT_DONE;
	     io_buffer_drainpacket(tmpbuf)) {

		if (io_buffer_extend(mybuf, len) == -1)
			return (IOFILT_ERROR);

		if (len < SHA_DIGEST_LENGTH)
			return (IOFILT_ERROR);

		ctx = auth->authin;

		HMAC_Update(&ctx, tmpbuf->data,
			    len - SHA_DIGEST_LENGTH + sizeof(u_int32_t));

		dlen = SHA_DIGEST_LENGTH;
		HMAC_Final(&ctx, digest, &dlen);

		if (memcmp(data + len - SHA_DIGEST_LENGTH, digest,
			   SHA_DIGEST_LENGTH)) {
			fprintf(stderr, __FUNCTION__": authentication failed");
			return (IOFILT_ERROR);
		}

		len -= SHA_DIGEST_LENGTH;
		if (len % 8) {
			fprintf(stderr, __FUNCTION__": bad length: %d", len);
			return (IOFILT_ERROR);
		}

		memcpy(iv, auth->iniv, 8);
		memcpy(auth->iniv, data + len - 8, 8);
		BF_cbc_encrypt(data, mybuf->data + mybuf->off,
			       len, &auth->in, iv, 0);

		data = mybuf->data + mybuf->off + len - 1;

		i = *data;
		if (i > 8)
			return (IOFILT_ERROR);

		len -= (i + 1);

		while (*data && *data == i) {
			data--;
			i--;
		}

		mybuf->off += len;
	}

	if (mybuf->off) {
		mybuf->ready = mybuf->off;
		return (IOFILT_DONE);
	}

	return (IOFILT_NEEDMORE);
}

void
crypto_addrandom(void)
{
	u_int32_t tmp[512];
	int i;

	/* Add some randomness */
	for (i = 0; i < 512; i++)
		tmp[i] = arc4random();
	RAND_add(tmp, sizeof(tmp), sizeof(tmp));
}
