/*
 * Copyright 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <libgen.h>

#include "config.h"

#include <openssl/sha.h>

#include "dataset.h"

int
dataset_readfile(struct dataset *sf, char *filename)
{
	u_char block[4096];
	struct stat sb;
	int fd;
	ssize_t n = -1, all;
	SHA_CTX ctx;

	if (stat(filename, &sb) == -1 || !(sb.st_mode & S_IFREG))
		return (-1);

	fd = open(filename, O_RDONLY, 0);
	if (fd == -1)
		return (-1);

	if ((all = sb.st_size) == 0)
		return (-1);

	SHA1_Init(&ctx);
	while (all && (n = read(fd, block, all > sizeof(block)
				? sizeof(block) : all)) != -1) {
		SHA1_Update(&ctx, block, n);
		all -= n;
	}

	close(fd);

	if (n == -1)
		return (-1);

	/* Put the data into the schedule */
	memset(sf, 0, sizeof(struct dataset));
	strlcpy(sf->sf_dir, dirname(filename), sizeof(sf->sf_name));
	strlcpy(sf->sf_name, basename(filename), sizeof(sf->sf_name));
	sf->sf_size = sb.st_size;
	SHA1_Final(sf->sf_hash, &ctx);

	return (0);
}

int
dataset_verify(struct dataset *sf, struct dataset *sftwo)
{
	size_t size;
	size = sizeof(sf->sf_name) + sizeof(sf->sf_hash) + sizeof(sf->sf_size);

	return (memcmp(&sftwo->sf_name, &sf->sf_name, size) == 0);
}

/* Verify if the file is still the same, XXX - slow? */
int
dataset_verifyfile(struct dataset *sf, char *filename)
{
	struct dataset tmpsf;

	if (dataset_readfile(&tmpsf, filename) == -1)
		return (0);

	return (dataset_verify(sf, &tmpsf));
}

void
dataset_deletefile(struct dataset *sf, struct dataset *tmpsf, int *pnfiles)
{
	int nfiles = *pnfiles;
	int offset = (tmpsf - sf) / sizeof(struct dataset);

	memmove(sf, sf + 1, (nfiles - offset - 1) * sizeof(struct dataset));
	nfiles--;

	*pnfiles = nfiles;
}

struct dataset *
dataset_find(struct dataset *sf, int nfiles, char *job, char *filename)
{
	char *normal = basename(filename);
	int i;

	for (i = 0; i < nfiles; i++, sf++)
		if ((job == NULL || strcmp(job, sf->sf_job) == 0) &&
		    strcmp(normal, sf->sf_name) == 0)
			break;
	if (i >= nfiles)
		return (NULL);

	return (sf);
}

/* Finds a file and verify if it is still there */

struct dataset *
dataset_findfile(struct dataset *sf, int *pnfiles, char *job, char *filename)
{
	int nfiles = *pnfiles;
	struct dataset *tmpsf;

	if ((tmpsf = dataset_find(sf, nfiles, job, filename)) == NULL)
		return (NULL);

	if (!dataset_verifyfile(tmpsf, filename)) {
		dataset_deletefile(sf, tmpsf, pnfiles);
		return (NULL);
	}

	return (sf);
}

struct dataset *
dataset_addfile(struct dataset *sf, int *pnfiles, char *filename, char *job)
{
	int nfiles = *pnfiles;

	if (nfiles >= MAX_DATASETS)
		return (NULL);

	sf += nfiles;

	if (dataset_readfile(sf, filename) == -1)
		return (NULL);

	if (job != NULL)
		strlcpy(sf->sf_job, job, sizeof(sf->sf_job));

	nfiles++;

	*pnfiles = nfiles;
	return (sf);
}

void
dataset_print(char *buf, size_t size, struct dataset *sf, int nfiles)
{
	char tmp[256];
	int i;

	snprintf(tmp, sizeof(tmp), "\t%d file%s:\n",
		 nfiles, nfiles > 1 ? "s" : "");
	strlcat(buf, tmp, size);

	for (i = 0; i < nfiles; i++) {
		snprintf(tmp, sizeof(tmp),
			 "\t\t%s%s%s/%s (%lld)\n",
			 sf->sf_job,
			 strlen(sf->sf_job) ? ": " : "",
			 sf->sf_dir,
			 sf->sf_name, sf->sf_size);
		strlcat(buf, tmp, size);

		sf++;
	}
}
