/*
 * Copyright 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/time.h>
#include <err.h>

#include <netinet/in.h>

#include "config.h"

#include <openssl/bn.h>

#include <event.h>
#include <io.h>

#include "io_aux.h"

int
io_filter_packet(struct io_filter *filt, void *data, size_t len)
{
	struct io_buffer buf;
	int res;

	memset(&buf, 0, sizeof(buf));
	io_buffer_packet(&buf, data, len);
	buf.flags |= IOBUFFLAG_FAST;
	buf.ready = buf.off;

	res = io_buffer_write((struct io_obj *)filt, &buf);
	io_buffer_free(&buf);

	return (res);
}

int
io_buffer_packet(struct io_buffer *buf, void *data, size_t len)
{
	u_int32_t nlen;
	int size = len + sizeof(nlen);

	if (io_buffer_extend(buf, size) == -1)
		errx(1, __FUNCTION__": io_buffer_extend");

	nlen = htonl(len);
	memcpy(buf->data + buf->off, &nlen, sizeof(nlen));
	memcpy(buf->data + buf->off + sizeof(nlen), data, len);

	buf->off += size;

	return (0);
}

int
io_buffer_bignum(struct io_buffer *buf, BIGNUM *n)
{
	u_int32_t blen, bytes = BN_num_bytes(n);
	int size = bytes + sizeof(bytes);

	if (io_buffer_extend(buf, size) == -1)
		errx(1, __FUNCTION__": io_buffer_extend");

	blen = BN_bn2bin(n, buf->data + buf->off + sizeof(bytes));
	if (blen != bytes)
		errx(1, __FUNCTION__": BN_bn2bin");

	bytes = htonl(bytes);
	memcpy(buf->data + buf->off, &bytes, sizeof(u_int32_t));

	buf->off += size;

	return (0);
}

int
io_buffer_haspacket(struct io_buffer *buf, void **pdata, size_t *plen)
{
	u_int32_t len;

	if (buf->off < sizeof(u_int32_t))
		return (IOFILT_NEEDMORE);

	memcpy(&len, buf->data, sizeof(u_int32_t));
	len = ntohl(len);

	if (buf->off < sizeof(u_int32_t) + len)
		return (IOFILT_NEEDMORE);

	if (pdata != NULL)
		*pdata = buf->data + sizeof(u_int32_t);
	if (plen != NULL)
		*plen = len;

	return (IOFILT_DONE);
}

void
io_buffer_drainpacket(struct io_buffer *buf)
{
	void *raw;
	u_int32_t rlen;

	if (io_buffer_haspacket(buf, &raw, &rlen) != IOFILT_DONE)
		return;

	io_buffer_drain(buf, rlen + sizeof(u_int32_t));
}

int
io_buffer_readpacket(struct io_buffer *buf, void *data, size_t len)
{
	void *raw;
	u_int32_t rlen;

	if (io_buffer_haspacket(buf, &raw, &rlen) == IOFILT_NEEDMORE)
		return (IOFILT_NEEDMORE);

	if (rlen != len)
		return (IOFILT_ERROR);

	memcpy(data, raw, len);

	io_buffer_drain(buf, rlen + sizeof(u_int32_t));

	return (IOFILT_DONE);
}

int
io_buffer_getbignum(struct io_buffer *buf, BIGNUM **pn)
{
	void *raw;
	u_int32_t rlen;
	BIGNUM *n;

	if (io_buffer_haspacket(buf, &raw, &rlen) == IOFILT_NEEDMORE)
		return (IOFILT_NEEDMORE);

	n = BN_bin2bn(raw, rlen, NULL);
	if (n == NULL)
		return (IOFILT_ERROR);

	io_buffer_drain(buf, rlen + sizeof(u_int32_t));

	*pn = n;

	return (IOFILT_DONE);
}
