/*
 * Copyright 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <libgen.h>
#include <errno.h>

#include "config.h"

#include <event.h>
#include <io.h>

#include "job.h"
#include "client.h"
#include "commands.h"
#include "util.h"

TAILQ_HEAD(joblist, job) jobqueue;

int job_signalcb(void);
void job_signalhandler(int);

int sig_chld = 0;

extern int (*event_sigcb)(void);
extern int event_gotsig;

int job_nice = 10;

void
job_init(void)
{
	TAILQ_INIT(&jobqueue);

	if (mysignal(SIGCHLD, job_signalhandler) == SIG_ERR)
		err(1, "signal");

	event_sigcb = job_signalcb;
}

void
job_signalhandler(int signal)
{
	event_gotsig = 1;

	if (signal == SIGCHLD) {
		mysignal(SIGCHLD, job_signalhandler);
		sig_chld = 1;
	}
}

struct job *
job_new(char *id)
{
	struct job *job;

	job = calloc(1, sizeof(struct job));
	if (job == NULL) {
		warn("calloc");
		return (NULL);
	}

	if ((job->job_id = strdup(id)) == NULL) {
		warn("strdup");
		free(job);
		return (NULL);
	}

	TAILQ_INSERT_TAIL(&jobqueue, job, job_next);

	return (job);
}

void
job_free_io(struct job *job)
{
	if (job->job_filterstdout != NULL) {
		io_terminate((struct io_obj *)job->job_filterstdout,
			     IORES_END);
		job->job_filterstdout = NULL;
	}
	if (job->job_filterstderr != NULL) {
		io_terminate((struct io_obj *)job->job_filterstderr,
			     IORES_END);
		job->job_filterstderr = NULL;
	}
}

void
job_free(struct job *job)
{
	TAILQ_REMOVE(&jobqueue, job, job_next);

	if (job->job_id != NULL)
		free(job->job_id);

	job_free_io(job);

	free(job);
}

struct job *
job_find(char *id)
{
	struct job *job;

	TAILQ_FOREACH(job, &jobqueue, job_next) {
		if (strcasecmp(job->job_id, id) == 0) {
			TAILQ_REMOVE(&jobqueue, job, job_next);
			TAILQ_INSERT_HEAD(&jobqueue, job, job_next);

			return (job);
		}
	}
	
	return (NULL);
}

struct job *
job_find_pid(pid_t pid)
{
	struct job *job;

	TAILQ_FOREACH(job, &jobqueue, job_next) {
		if (job->job_pid == pid)
			return (job);
	}
	
	return (NULL);
}

int
job_execute_cmd(char *job, char *cmd, char **argv)
{
	pid_t pid;
	int status, res, save_errno = 0;
	sigset_t sigmask;

	/* Block SIGCHLD */
	sigemptyset(&sigmask);
	sigaddset(&sigmask, SIGCHLD);
	if (sigprocmask(SIG_BLOCK, &sigmask, NULL) == -1) {
		warn("sigprocmask");
		return (-1);
	}

	pid = fork();
	if (pid == -1) {
		warn("fork");
		return (-1);
	}

	if (pid == 0) {
		/* Child */
		if (chdir(job) == -1)
			err(1, __FUNCTION__": chdir");

		if (execv(cmd, argv) == -1)
			err(1, __FUNCTION__": execv(%s)", cmd);

		/* NOT REACHED */
	}

	/* waitpid should never be interrupted by a signal, but maybe it
	 * can on some systems. */
	while ((res = waitpid(pid, &status, 0)) == -1) {
		if (errno != EINTR) {
			save_errno = errno;
			break;
		}
	}

	/* Install old signal handler */
	if (sigprocmask(SIG_UNBLOCK, &sigmask, NULL) == -1) {
		warn("sigprocmask");
		return (-1);
	}

	if (res == -1) {
		errno = save_errno;
		warn("waitpid");
		return (-1);
	}

	if (!WIFEXITED(status)) {
		warnx(__FUNCTION__": didn't exit");
		return (-1);
	}

	return (WEXITSTATUS(status) == 0 ? 0 : -1);
}

char *
job_statusline(void)
{
	static char status[1024];
	char tmp[128];
	struct job *job;

	status[0] = '\0';

	TAILQ_FOREACH(job, &jobqueue, job_next) {
		snprintf(tmp, sizeof(tmp), "%s%s(%d)",
			 *status != '\0' ? ", " : "",
			 job->job_id, job->job_pid);
		strlcat(status, tmp, sizeof(status));
	}

	return (status);
}

/* Communicate job output back to server */
void
filter_jobiodead(struct io_obj *this, void *arg)
{
	struct io_filter **filt = arg;

	/* This is pretty evil */
	*filt = NULL;
}

int
filter_combine(struct job *job, char which, struct io_filter *filter,
	       struct io_buffer *buf)
{
	struct io_buffer *mybuf = &filter->io_buf;
	struct io_buffer *tmpbuf = &filter->io_tmpbuf;
	char line[128];
	int i;

	if (buf->ready == 0)
		return (IOFILT_NEEDMORE);

	if (io_buffer_append(tmpbuf, buf) == -1)
		return (IOFILT_ERROR);

	if (memchr(tmpbuf->data, '\0', tmpbuf->off) != NULL) {
		/* This buffer contains binary data */
		snprintf(line, sizeof(line), "%s %c %s: ",
			 CMD_OUTPUT, which, job->job_id);
		io_buffer_data(mybuf, line, strlen(line));

		tmpbuf->ready = tmpbuf->off;
		goto done;
	}

	/* Do we have a whole line */
	if (memchr(tmpbuf->data, '\n', tmpbuf->off) == NULL)
		return (IOFILT_NEEDMORE);

	snprintf(line, sizeof(line), "%s %c %s: ",
		 CMD_OUTPUT, which, job->job_id);
	io_buffer_data(mybuf, line, strlen(line));

	/* Deal with lines, there is a '\n' in here */
	for (i = tmpbuf->off - 1; i >= 0; i--)
		if (tmpbuf->data[i] == '\n')
			break;

	tmpbuf->ready = i + 1;

 done:
	io_buffer_append(mybuf, tmpbuf);
	io_buffer_drain(tmpbuf, tmpbuf->ready);

	io_buffer_data(mybuf, "\0", 1);
	client_bufferorsend(mybuf->data, mybuf->off);
	io_buffer_drain(mybuf, mybuf->off);

	return (IOFILT_NEEDMORE);
}

int
filter_stdout(void *arg, struct io_filter *filter,
	      struct io_buffer *buf, int flags)
{
	struct job *job = arg;

	return (filter_combine(job, 'O', filter, buf));
}

int
filter_stderr(void *arg, struct io_filter *filter,
	      struct io_buffer *buf, int flags)
{
	struct job *job = arg;

	return (filter_combine(job, 'E', filter, buf));
}

int
job_fork_cmd(struct job *job, char *cmd, char **argv)
{
	struct io_obj *source;
	int pout[2], perr[2];

	job->job_filterstdout = io_new_filter(filter_stdout, job);
	job->job_filterstderr = io_new_filter(filter_stderr, job);
	if (job->job_filterstdout == NULL || job->job_filterstderr == NULL)
		goto out_err;

	if (pipe(pout) == -1) {
		warn("pipe");
		goto out_err;
	}

	if (pipe(perr) == -1) {
		warn("pipe");
		goto pipe_err;
	}

	job->job_pid = fork();
	if (job->job_pid == -1) {
		warn("fork");
		goto fork_err;
	}

	if (job->job_pid == 0) {
		/* Child */
		if (dup2(pout[1], fileno(stdout)) == -1)
			err(1, __FUNCTION__": dup2");
		if (dup2(perr[1], fileno(stderr)) == -1)
			err(1, __FUNCTION__": dup2");

		close(pout[0]);
		close(perr[0]);

		if (job_nice)
			setpriority(PRIO_PROCESS, 0, job_nice);

		if (chdir(job->job_id) == -1)
			err(1, __FUNCTION__": chdir");

		if (execv(cmd, argv) == -1)
			err(1, __FUNCTION__": execv(%s)", cmd);

		/* NOT REACHED */
	}

	/* Let's hope that none of these fail - XXX */
	source = io_new_source(pout[0]);
	source->io_free = filter_jobiodead;
	source->io_freearg = &job->job_filterstdout;
	io_attach(source, job->job_filterstdout);

	source = io_new_source(perr[0]);
	source->io_free = filter_jobiodead;
	source->io_freearg = &job->job_filterstderr;
	io_attach(source, job->job_filterstderr);

	/* Close write side */
	close(pout[1]);
	close(perr[1]);

	return (0);

	/* Error cleanup */
 fork_err:
	close(perr[0]);
	close(perr[1]);
 pipe_err:
	close(pout[0]);
	close(pout[1]);
 out_err:
	job->job_pid = 0;
	job_free_io(job);
	return (-1);
}

int
job_signalcb(void)
{
	struct job *job;
	int status, died;
	pid_t pid;

	if (!sig_chld)
		return (0);

	sig_chld = 0;

	do {
#ifdef HAVE_WAIT3
		pid = wait3(&status, WNOHANG, NULL);
#else
		pid = wait4(-1, &status, WNOHANG, NULL);
#endif

		if (pid == -1) {
			if (errno == ECHILD)
				break;
			if (errno == EINTR)
				continue;
			/* This is a real error now */
			return (0);
		}

		/* Some operating systems don't know about POSIX */
		if (WIFSTOPPED(status)) {
			fprintf(stderr, __FUNCTION__": child %d: stopped %x\n",
			    pid, status);
			continue;
		}

		if (!WIFEXITED(status) && !WIFSIGNALED(status)) {
			fprintf(stderr, __FUNCTION__": child %d: %x\n",
				pid, status);
			continue;
		}

		if (WIFEXITED(status))
			died = WEXITSTATUS(status);
		else
			died = 1;

		job = job_find_pid(pid);
		if (job == NULL) {
			fprintf(stderr, __FUNCTION__": no info for pid %d\n",
				pid);
			continue;
		}

		/* Clean up */
		job->job_pid = 0;

		client_jobresult(job, died);
	} while (pid > 0);

	return (0);
}
