/*
 * Copyright 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <libgen.h>
#include <errno.h>

#include <openssl/rsa.h>
#include <openssl/dh.h>

#include "config.h"

#include <event.h>
#include <io.h>

#include "schedule.h"
#include "auth.h"
#include "server.h"
#include "util.h"

TAILQ_HEAD(schedlist, schedule) schedqueue;

/* Parameters that might be configurable */
int slot_timeout = SLOT_TIMEOUT;

void
schedule_init(void)
{
	TAILQ_INIT(&schedqueue);
}

int
schedule_file(char *id, char *append)
{
	char name[1024];
	time_t now;
	int fd;

	strlcpy(name, id, sizeof(name));
	strlcat(name, append, sizeof(name));
	fd = open(name, O_WRONLY|O_CREAT|O_APPEND, 0644);
	if (fd == -1)
		warn("open(%s)", name);

	now = time(NULL);
	snprintf(name, sizeof(name), "Started %s", ctime(&now));
	write(fd, name, strlen(name));

	return (fd);
}

struct schedule *
schedule_new(char *id)
{
	struct schedule *sd;

	sd = calloc(1, sizeof(struct schedule));
	if (sd == NULL) {
		warn("calloc");
		return (NULL);
	}

	if ((sd->sd_id = strdup(id)) == NULL) {
		warn("strdup");
		goto out;
	}

	sd->sd_start = sd->sd_end = -1;

	/* Output files */
	if ((sd->sd_stdout = schedule_file(id, ".stdout")) == -1)
		goto out;
	if ((sd->sd_stderr = schedule_file(id, ".stderr")) == -1)
		goto out;

	TAILQ_INIT(&sd->sd_slqueue);

	TAILQ_INSERT_TAIL(&schedqueue, sd, sd_next);

	return (sd);
 out:
	if (sd->sd_id != NULL)
		free(sd->sd_id);
	free(sd);

	return (NULL);
}
 
void
schedule_free(struct schedule *sd)
{
	struct slot *slot;
	TAILQ_REMOVE(&schedqueue, sd, sd_next);

	/* Free up all the slots */
	for(slot = TAILQ_FIRST(&sd->sd_slqueue); slot;
	    slot = TAILQ_FIRST(&sd->sd_slqueue))
		slot_free(slot);

	if (sd->sd_stdout != -1)
		close(sd->sd_stdout);
	if (sd->sd_stderr != -1)
		close(sd->sd_stderr);

	if (sd->sd_id != NULL)
		free(sd->sd_id);
	if (sd->sd_cmd != NULL)
		free(sd->sd_cmd);
	if (sd->sd_bitmap != NULL);
		free(sd->sd_bitmap);

	free(sd);
}

struct schedule *
schedule_find(char *id)
{
	struct schedule *schedule;

	TAILQ_FOREACH(schedule, &schedqueue, sd_next) {
		if (strcasecmp(schedule->sd_id, id) == 0) {
			TAILQ_REMOVE(&schedqueue, schedule, sd_next);
			TAILQ_INSERT_HEAD(&schedqueue, schedule, sd_next);

			return (schedule);
		}
	}
	
	return (NULL);
}

void
schedule_update(struct schedule *sd)
{
	size_t size;

	if (!SD_HAS_RANGE(sd))
		return;

	size = sd->sd_end - sd->sd_start + 1;
	sd->sd_left = size;

	if (sd->sd_bitmap != NULL)
		free(sd->sd_bitmap);

	/* Allocate a bitstring to keep track of things we hand out */
	sd->sd_nbits = size;
	sd->sd_bitmap = bit_alloc(sd->sd_nbits);
	if (sd->sd_bitmap == NULL)
		err(1, __FUNCTION__": bit_alloc");
	bit_nclear(sd->sd_bitmap, 0, sd->sd_nbits - 1);
}

char *
schedule_status(struct schedule *sched, int verbose)
{
	static char status[4096];
	char tmp[128];
	
	status[0] = '\0';
	snprintf(tmp, sizeof(tmp), "Schedule: %s\n", sched->sd_id);
	strlcat(status, tmp, sizeof(status));
	if (sched->sd_flags) {
		strlcpy(tmp, "\tFlags:", sizeof(tmp));
		if (sched->sd_flags & SDFLAGS_AUTOMATIC)
			strlcat(tmp, " Automatic", sizeof(tmp));
		if (sched->sd_flags & SDFLAGS_LOCKED)
			strlcat(tmp, " Locked", sizeof(tmp));

		strlcat(status, tmp, sizeof(status));
		strlcat(status, "\n", sizeof(status));
	}

	if (sched->sd_nfiles) {
		dataset_print(status, sizeof(status), sched->sd_files,
			      sched->sd_nfiles);
	}

	snprintf(tmp, sizeof(tmp),
		 "\tRemote command:  %s\n",
		 sched->sd_cmd != NULL ? sched->sd_cmd : "(unset)");
	strlcat(status, tmp, sizeof(status));

	if (sched->sd_start != -1 && sched->sd_end != -1)
		snprintf(tmp, sizeof(tmp),
			 "\tParameter range: %d-%d\n",
			 sched->sd_start, sched->sd_end);
	else
		snprintf(tmp, sizeof(tmp), "\tParameter range: (unset)\n");
	strlcat(status, tmp, sizeof(status));

	snprintf(tmp, sizeof(tmp),
		 "\tSlots assigned:  %d\n"
		 "\tSlots finished:  %d\n"
		 "\tSlots left:      %d\n",
		 sched->sd_assigned,
		 sched->sd_done,
		 sched->sd_left);
	strlcat(status, tmp, sizeof(status));
	if (sched->sd_orphaned) {
		snprintf(tmp, sizeof(tmp), "\tSlots orphaned:  %d\n",
		    sched->sd_orphaned);
		strlcat(status, tmp, sizeof(status));
	}

	if (verbose && sched->sd_assigned) {
		time_t now = time(NULL);
		struct slot *slot;
		strlcat(status, "Assigned slots:\n", sizeof(status));

		TAILQ_FOREACH(slot, &sched->sd_slqueue, sl_next) {
			snprintf(tmp, sizeof(tmp),
			    "\t%d -> %s:%u%s, %s\n",
			    slot->sl_slot,
			    slot->sl_clname, slot->sl_clport,
			    slot->sl_flags & SLFLAGS_CLIENTGONE
			    ? " (orphaned)" : "",
			    elapsedtime(now - slot->sl_when));
			strlcat(status, tmp, sizeof(status));
		}
	}

	if (sched->sd_done) {
		snprintf(tmp, sizeof(tmp), "\tTime Min: %s, ",
			 elapsedtime(sched->sd_timemin));
		strlcat(status, tmp, sizeof(status));
		snprintf(tmp, sizeof(tmp), "Avg: %s, ",
			 elapsedtime(sched->sd_timecount / sched->sd_done));
		strlcat(status, tmp, sizeof(status));
		snprintf(tmp, sizeof(tmp), "Max: %s\n",
			 elapsedtime(sched->sd_timemax));
		strlcat(status, tmp, sizeof(status));
	}

	if (sched->sd_timestart) {
		time_t now;

		if (sched->sd_timeend == 0)
			now = time(NULL) - sched->sd_timestart;
		else
			now = sched->sd_timeend - sched->sd_timestart;
		snprintf(tmp, sizeof(tmp),
			 "\tRunning time: %s, Slot-rate: %s\n",
			 elapsedtime(now),
			 sched->sd_done
			 ? time2rate((float)now/sched->sd_done) : "n.a.");
		strlcat(status, tmp, sizeof(status));
	}

	return (status);
}

void
schedule_statusline(char *data, size_t size, int i, struct schedule *tmp)
{
	char tmpbuf[128];

	snprintf(data, size, "%d. %s", i, tmp->sd_id);

	if (tmp->sd_cmd != NULL) {
		snprintf(tmpbuf, sizeof(tmpbuf), " - %s", tmp->sd_cmd);
		strlcat(data, tmpbuf, size);
	}

	strlcat(data, "\n", size);
}

#define MSG_NOSCHEDULE	"No jobs configured.\n"

void
schedule_list(struct io_buffer *buf)
{
	char line[1024];
	struct schedule *tmp;
	int i;

	if (TAILQ_FIRST(&schedqueue) == NULL) {
		io_buffer_data(buf, MSG_NOSCHEDULE, sizeof(MSG_NOSCHEDULE) - 1);
		return;
	}

	i = 1;
	TAILQ_FOREACH(tmp, &schedqueue, sd_next) {
		schedule_statusline(line, sizeof(line), i, tmp);
		io_buffer_data(buf, line, strlen(line));
		i++;
	}
}

int
schedule_getslot(struct schedule *sd)
{
	int slot;

	bit_ffc(sd->sd_bitmap, sd->sd_nbits, &slot);
	bit_set(sd->sd_bitmap, slot);

	slot += sd->sd_start;

	sd->sd_assigned++;
	sd->sd_left--;

	return (slot);
}

void
schedule_returnslot(struct schedule *sd, int slot)
{
	slot -= sd->sd_start;

	if (!bit_test(sd->sd_bitmap, slot))
		errx(1, __FUNCTION__": data corrupt");

	bit_clear(sd->sd_bitmap, slot);

	sd->sd_assigned--;
	sd->sd_left++;
}

void
schedule_slotdone(struct client *client, char *jobid, int died)
{
	struct schedule *sd;
	struct slot *slot;
	struct timeval tv;

	sd = schedule_find(jobid);
	if (sd == NULL) {
		fprintf(stderr, __FUNCTION__": %s not found\n", jobid);
		return;
	}

	if (!SD_HAS_RANGE(sd))
		return;

	slot = slot_find_id(sd, client->auth->client_id);
	if (slot == NULL) {
		fprintf(stderr, __FUNCTION__": %s no slot for client\n",
			jobid);
		return;
	}

	/* Prepare for delayed callback */
	timerclear(&tv);
	tv.tv_sec = SLOT_WAIT;

	if (died)
		slot->sl_flags |= SLFLAGS_DIED;

	timeout_add(&slot->sl_wait, &tv);
}

char *
schedule_expand(struct schedule *sd, char *cmd, int slot)
{
	static char expand[1024];

	snprintf(expand, sizeof(expand), "%s", cmd);

	while (strrpl(expand, sizeof(expand), "$slot", slot) != NULL)
		;

	if (sd != NULL) {
		while (strrpl(expand, sizeof(expand),
			   "$start", sd->sd_start) != NULL)
			;
		while (strrpl(expand, sizeof(expand),
			   "$end", sd->sd_end) != NULL)
			;
	}

	return (expand);
}

int
schedule_assign(struct schedule *sd, struct client *client, int *pslot)
{
	struct slot *slot;
	int slot_nr = -1;
	
	if (SD_HAS_RANGE(sd)) {
		/* See if there are any slots left */
		if (sd->sd_left == NULL)
			return (-1);

		slot = slot_new(sd);

		slot->sl_clname = strdup(client->name);
		if (slot->sl_clname == NULL) {
			warn(__FUNCTION__": strdup");
			/* Return it */
			schedule_returnslot(sd, slot->sl_slot);
			slot_free(slot);
			return (-1);
		}
		slot->sl_clport = client->port;

		/* Permanent Identifier */
		memcpy(slot->sl_clid, client->auth->client_id,
		       sizeof(slot->sl_clid));
		slot_nr = slot->sl_slot;
	}

	*pslot = slot_nr;
	return (0);
}

void
schedule_clientdie(struct client *client)
{
	struct schedule *sd;
	struct slot *slot;
	struct timeval tv;

	timerclear(&tv);
	tv.tv_sec = slot_timeout;
	TAILQ_FOREACH(sd, &schedqueue, sd_next) {
		if ((slot = slot_find_id(sd, client->auth->client_id)) == NULL)
			continue;

		sd->sd_orphaned++;
		slot->sl_flags |= SLFLAGS_CLIENTGONE;
		timeout_add(&slot->sl_timeout, &tv);
	}
}

/* When a client reconnects, have him reclaim his old slots */

void
schedule_clientconnect(struct client *client, char *job)
{
	struct schedule *sd;
	struct slot *slot;

	if ((sd = schedule_find(job)) == NULL)
		return;
	
	if ((slot = slot_find_id(sd, client->auth->client_id)) == NULL)
		return;
	if (!(slot->sl_flags & SLFLAGS_CLIENTGONE))
		return;

 	sd->sd_orphaned--;
	slot->sl_flags &= ~SLFLAGS_CLIENTGONE;

	fprintf(stderr, __FUNCTION__": %s:%u reclaims %s(%d)\n",
	    client->name, client->port, sd->sd_id, slot->sl_slot);
	timeout_del(&slot->sl_timeout);
}

/*
 * Manually reclaim orphaned slots. 
 */

void
schedule_reclaim(struct schedule *sd, struct io_buffer *buf)
{
	char line[1024];
	struct slot *slot, *next;

	/* Free up all the slots */
	for(slot = TAILQ_FIRST(&sd->sd_slqueue); slot; slot = next) {
		next = TAILQ_NEXT(slot, sl_next);

		if (!(slot->sl_flags & SLFLAGS_CLIENTGONE))
			continue;

		snprintf(line, sizeof(line), "%d -> %s:%u reclaimed\n",
		    slot->sl_slot, slot->sl_clname, slot->sl_clport);
		io_buffer_data(buf, line, strlen(line));

		/* Reclaim this orphaned slot */
		timeout_del(&slot->sl_timeout);
		slot_lostclient(-1, EV_TIMEOUT, slot);
	}
}

/* Called if we got a process result for a slot */

void
slot_result(struct schedule *sd, struct slot *slot, int died)
{
	int slotnr = slot->sl_slot;
	time_t duration, now = time(NULL);

	if (!sd->sd_timestart)
		sd->sd_timestart = slot->sl_when;

	duration = now - slot->sl_when;

	slot_free(slot);

	if (died) {
		schedule_returnslot(sd, slotnr);
	} else {
		sd->sd_timecount += duration;

		if (sd->sd_timemax < duration)
			sd->sd_timemax = duration;
		if (!sd->sd_done || sd->sd_timemin > duration)
			sd->sd_timemin = duration;

		sd->sd_assigned--;
		sd->sd_done++;

		if (sd->sd_assigned == 0 && sd->sd_left == 0)
			sd->sd_timeend = now;
	}
}

/* When a slot times out, reclaim it, so that it can be reassigned */

void
slot_lostclient(int fd, short which, void *arg)
{
	struct slot *slot = arg;
	struct schedule *sd = slot->sl_sd;

	/* Now return it */
	slot_result(sd, slot, 1);
}

/* We got a result for a slot, but delay acting upon it */

void
slot_waitclient(int fd, short which, void *arg)
{
	struct slot *slot = arg;
	struct schedule *sd = slot->sl_sd;
	struct client *client;
	int died = slot->sl_flags & SLFLAGS_DIED;

	if (died || !(sd->sd_flags & SDFLAGS_AUTOMATIC)) {
		slot_result(sd, slot, died);
		return;
	}

	/* Refind the same client, maybe we should reschedule based on
	 * the clients status line.
	 */
	client = client_find_id(slot->sl_clid);

	slot_result(sd, slot, died);

	/* See if we should start a new job on this client */
	if (client != NULL)
		client_run_one(client, sd);
}

struct slot *
slot_new(struct schedule *sd)
{
	struct slot *slot;

	slot = calloc(1, sizeof(struct slot));
	if (slot == NULL) {
		warn(__FUNCTION__": calloc");
		return (NULL);
	}

	slot->sl_slot = schedule_getslot(sd);
	slot->sl_when = time(NULL);

	slot->sl_sd = sd;

	timeout_set(&slot->sl_timeout, slot_lostclient, slot);
	timeout_set(&slot->sl_wait, slot_waitclient, slot);

	TAILQ_INSERT_TAIL(&sd->sd_slqueue, slot, sl_next);

	return (slot);
}

void
slot_free(struct slot *slot)
{
	if (slot->sl_flags & SLFLAGS_CLIENTGONE)
		slot->sl_sd->sd_orphaned--;
	
	TAILQ_REMOVE(&slot->sl_sd->sd_slqueue, slot, sl_next);

	timeout_del(&slot->sl_timeout);
	timeout_del(&slot->sl_wait);
	if (slot->sl_clname)
		free(slot->sl_clname);

	free(slot);
}

struct slot *
slot_find_id(struct schedule *sd, u_char *id)
{
	struct slot *slot;

	TAILQ_FOREACH(slot, &sd->sd_slqueue, sl_next) {
		if (memcmp(slot->sl_clid, id, sizeof(slot->sl_clid)) == 0)
			break;
	}

	return (slot);
}

struct slot *
slot_find_nr(struct schedule *sd, int nr)
{
	struct slot *slot;

	TAILQ_FOREACH(slot, &sd->sd_slqueue, sl_next) {
		if (slot->sl_slot == nr)
			break;
	}

	return (slot);
}
