/*
 * Copyright 2003 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <err.h>

#include <openssl/rsa.h>
#include <openssl/dh.h>

#include "config.h"

#include <event.h>
#include <io.h>

#include "crypto.h"
#include "auth.h"
#include "server.h"
#include "ui.h"
#include "schedule.h"

#define RSA_KEY_BITS	1024
#define RSA_KEY_E	37

char *progname;

/* Our private key */
RSA *srv_key;

struct io_obj *stdsrc, *stddst;
struct io_filter *anchor;

int
filter_clientcomplete(void *arg, struct io_filter *filter,
		      struct io_buffer *buf, int flags)
{
	struct client *client= arg;
	struct auth_state *auth = client->auth;
	struct io_filter *filt;

	io_detach(auth->dplx->io_dst, filter);

	filter = client->auth_filter;
	io_detach(auth->dplx->io_src, filter);
	io_detach(filter, auth->dplx->io_dst);

	filt = io_new_filter(filter_encrypt, auth);
	io_attach(filt, auth->dplx->io_dst);
	/* Protect encrypt filter */
	io_attach(anchor, filt);

	client->dst = filt;

	filt = io_new_filter(filter_decrypt, auth);
	io_attach(auth->dplx->io_src, filt);
	/* Protect decrypt filter */
	io_attach(filt, anchor);

	client->src = filt;

	filter = client->auth_filter;
	io_attach(client->src, filter);

	return (IOFILT_END);
}

int
filter_handleclient(void *arg, struct io_filter *filter,
		    struct io_buffer *buf, int flags)
{
	struct client *client = arg;
	struct auth_state *auth = client->auth;
	struct io_buffer *mybuf = &filter->io_buf;
	struct io_filter *filt;
	int res;

	if (io_buffer_append(mybuf, buf) == -1)
		return (-1);

	res = auth_protocol_server(mybuf, auth);

	if (auth->step == SAUTH_READCLIENTID) {
		filt = io_new_filter(filter_clientcomplete, client);
		io_attach(auth->dplx->io_dst, filt);
	} else if (auth->step == -1) {
		/* React to client reponses */
		filt = io_new_filter(filter_clientresponse, client);
		io_attach(client->src, filt);

		client->flags |= CLIENT_AUTHDONE;
		
		fprintf(stderr, __FUNCTION__
			": %s:%d server authentication complete\n",
			client->name, client->port);

		/* We might get status and info messages here XXX */
		if (mybuf->off) {
			mybuf->ready = mybuf->off;
			io_buffer_write((struct io_obj *)filt, mybuf);
		}
		
		return (IOFILT_END);
	}

	return (res);
}

int
server_authenticate(struct client *client)
{
	struct io_filter *filt;
	struct io_duplex *dplx;
	struct auth_state *auth;

	if ((auth = calloc(1, sizeof(struct auth_state))) == NULL) {
		warn(__FUNCTION__": calloc");
		return (-1);
	}

	client->auth = auth;
	dplx = client->dplx;

	/* Set a write timeout */
	io_set_timeout(dplx->io_dst, CLIENT_WRITETIMEOUT);

	auth->rsa_key = srv_key;
	auth->dplx = dplx;

	filt = io_new_filter(filter_handleclient, client);
	io_attach(dplx->io_src, filt);
	io_attach(filt, dplx->io_dst);

	client->auth_filter = filt;

	/* Only set this if nothing below can fail */
	dplx->io_src->io_free = client_dead;
	dplx->io_src->io_freearg = client;

	/* Start the auth protocol */
	auth_protocol_server(&filt->io_buf, auth);
	io_process_filter((struct io_obj *)filt, &filt->io_buf, 0);

	return (0);
}

int
client_connect(char *name, short port)
{
	struct client *client;
	
	if ((client = client_new(name, port)) == NULL) {
		warnx(__FUNCTION__": client_new(%s, %d)", name, port);
		return (-1);
	}

	client->dplx = io_new_connector(name, port);
	if (client->dplx == NULL) {
		warnx(__FUNCTION__": io_new_connector(%s, %d)", name, port);
		return (-1);
	}

	if (server_authenticate(client) == -1) {
		io_terminate(client->dplx->io_src, IORES_ERROR);
		client_free(client);
	}

	return (0);
}

int
filter_handlenew(void *arg, struct io_filter *filter, struct io_buffer *buf, 
		 int flags)
{
	struct client *client = NULL;
	struct io_buffer *mybuf = &filter->io_buf;
	struct io_duplex *dplx = NULL;
	struct sockaddr_storage *from;
	size_t fromlen;
	int newfd = -1;
	u_char *p;
	char name[NI_MAXHOST], port[NI_MAXSERV];

	if (buf->ready < sizeof (int) + sizeof(fromlen))
		return (IOFILT_ERROR);

	p = buf->data;
	newfd = *(int *)p;
	p += sizeof(int);
	fromlen = *(size_t *)p;
	p += sizeof(size_t);
	from = (struct sockaddr_storage *)p;

	/* We need to get the address */
	if (buf->ready < sizeof (int) + sizeof(fromlen) + fromlen)
		errx(1, __FUNCTION__": not enough data");

	if (getnameinfo((struct sockaddr *)from, fromlen,
			name, sizeof(name), port, sizeof(port),
			NI_NUMERICHOST | NI_NUMERICSERV) != 0) {
		warnx(__FUNCTION__": getnameinfo failed");
		goto error;
	}

	if (!strlen(name)) {
		warnx(__FUNCTION__": didn't get client address");
		goto error;
	}

	if (io_buffer_extend(mybuf, 256) == -1)
		goto error;

	client = client_new(name, atoi(port));
	if (client == NULL)
		goto error;

	if ((dplx = io_new_socket(newfd)) == NULL)
		goto error;
	client->dplx = dplx;

	if (server_authenticate(client) == -1)
		goto error;

	snprintf(mybuf->data, 256, __FUNCTION__
		 ": new connection on fd %d from %s:%s\n",
		 newfd, name, port);
	mybuf->ready = mybuf->off = strlen(mybuf->data);

	return (IOFILT_DONE);

 error:
	if (dplx != NULL)
		io_terminate(dplx->io_src, IORES_ERROR);
	else if (newfd != -1)
		close(newfd);

	if (client != NULL)
		client_free(client);

	return (IOFILT_NEEDMORE);
}

int
server_readclientlist(void)
{
	/* client_connect("127.0.0.1", 8080); */

	return (0);
}

void
usage(void)
{
	fprintf(stderr, "%s: [-f fifo]\n", progname);
	exit(1);
}

int
main(int argc, char **argv)
{
	extern char *optarg;
	extern int optind;
	extern char *ui_file;
	struct rlimit rl;
	char *keyname = "server-key-rsa";
	struct io_obj *fdsrc;
	struct io_filter *filt;
	int ch;
	
	progname = argv[0];

	while ((ch = getopt(argc, argv, "f:")) != -1) {
		switch (ch) {
		case 'f':
			ui_file = optarg;
			break;
			    
		default:
			usage();
		}
	}

	argc -= optind;
	argv += optind;

	crypto_addrandom();
	
	srv_key = rsa_read_key(keyname, RSA_PRIVATE);
	if (srv_key == NULL) {
		char tmpname[1024];

		fprintf(stderr, __FUNCTION__": Generating RSA key...\n");
		srv_key = RSA_generate_key(RSA_KEY_BITS, RSA_KEY_E,
					       NULL, NULL);
		if (srv_key == NULL)
			errx(1, "RSA_generate_key");

		rsa_write_key(keyname, srv_key, RSA_PRIVATE);

		strlcpy(tmpname, keyname, sizeof (tmpname));
		strlcat(tmpname, ".pub", sizeof (tmpname));
		rsa_write_key(tmpname, srv_key, RSA_PUBLIC);
	}

	/* Raising file descriptor limits */
	if (getrlimit(RLIMIT_NOFILE, &rl) == -1)
		err(1, "getrlimit");
	rl.rlim_cur=rl.rlim_max;
	if (setrlimit(RLIMIT_NOFILE, &rl) == -1)
		err(1, "setrlimit");

	event_init();
	schedule_init();

	anchor = io_new_filter(NULL, NULL);

	stdsrc = io_new_source(fileno(stdin));
	stddst = io_new_sink(fileno(stdout));

	/* Protect stdin and stdout */
	io_attach(stdsrc, anchor);
	io_attach(anchor, stddst);

	if ((fdsrc = io_new_listener("0.0.0.0", 8010)) == NULL)
		errx(1, __FUNCTION__": io_new_listener failed");

	filt = io_new_filter(filter_handlenew, NULL);
	io_attach(fdsrc, filt);
	io_attach(filt, stddst);

	/* Need anchor for ui_init */
	ui_init();
	client_init();

	server_readclientlist();

	event_dispatch();
	
	exit(0);
}
