#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <regex.h>

#include <openssl/rsa.h>
#include <openssl/dh.h>

#include "config.h"

#include <event.h>
#include <io.h>

#include "crypto.h"
#include "auth.h"
#include "server.h"
#include "io_aux.h"
#include "commands.h"
#include "schedule.h"
#include "ui.h"
#include "util.h"

struct clientlist clientqueue;
struct clientlist cmdqueue;

time_t client_now;

/* stats */
int client_errors = 0;
int client_done = 0;
int client_connected = 0;
int client_selected = 0;

struct io_buffer client_errbuf;
struct io_buffer client_donebuf;

void
client_init(void)
{
	TAILQ_INIT(&clientqueue);
	TAILQ_INIT(&cmdqueue);

	memset(&client_errbuf, 0, sizeof(client_errbuf));
	memset(&client_donebuf, 0, sizeof(client_donebuf));
}

void
client_dead(struct io_obj *obj, void *arg)
{
	struct client *client = arg;

	fprintf(stderr, __FUNCTION__": %s:%d died: %p\n",
		client->name, client->port, arg);

	client_free(client);
}

void
client_timeout(int fd, short which, void *arg)
{
	struct client *client = arg;

	fprintf(stderr, __FUNCTION__": connection to %s:%u timed out\n",
		client->name, client->port);

	io_terminate((struct io_obj *)client->src, IORES_ERROR);
}

struct client *
client_new(char *name, u_short port)
{
	struct client *client;

	client = calloc(1, sizeof(struct client));
	if (client == NULL) {
		warn(__FUNCTION__": calloc");
		return (NULL);
	}

	client->name = strdup(name);
	if (client->name == NULL) {
		free(client);
		return (NULL);
	}
	client->port = port;
	client->when = time(NULL);

	TAILQ_INIT(&client->client_cmdqueue);

	timeout_set(&client->ev, client_timeout, client);
	
	TAILQ_INSERT_TAIL(&clientqueue, client, client_next);

	client_connected++;
	
	return(client);
}

void
client_settimeout(struct client *client)
{
	struct timeval tv;

	timerclear(&tv);
	tv.tv_sec = CLIENT_WRITETIMEOUT;

	timeout_add(&client->ev, &tv);
}

void
client_free(struct client *client)
{
	struct client_cmd *cc;
	
	client->flags |= CLIENT_DEAD;

	for(cc = TAILQ_FIRST(&client->client_cmdqueue); cc;
	    cc = TAILQ_FIRST(&client->client_cmdqueue))
		command_free(client, cc);

	timeout_del(&client->ev);

	schedule_clientdie(client);

	TAILQ_REMOVE(&clientqueue, client, client_next);

	if (client->flags & CLIENT_ONCMDLIST) {
		TAILQ_REMOVE(&cmdqueue, client, cmd_next);
		client_selected--;
	}

	if (client->name != NULL)
		free(client->name);

	if (client->auth != NULL)
		auth_free_state(client->auth);

	client_connected--;
	
	free(client);
}

void
client_statusline(char *line, size_t size, int i, struct client *client)
{
	char *status;
	struct client_cmd *cc;
	int ncommands;

	ncommands = 0;
	TAILQ_FOREACH(cc, &client->client_cmdqueue, cc_next) {
		ncommands++;
	}

	if (client->flags & CLIENT_ONCMDLIST) {
		if (client->flags & CLIENT_CMDINPROGRESS)
			status = " [.]";
		else
			status = " (*)";
	} else
		status = "";

	snprintf(line, size,
		 "%d. %s:%u%s, %s%s%d cmds, connected for %s\n",
		 i,
		 client->name, client->port,
		 status,
		 client->status,
		 *client->status != '\0' ? ", " : "",
		 ncommands,
		 elapsedtime(client_now - client->when));
}

/* Find a client based on its unique id */

struct client *
client_find_id(u_char *id)
{
	struct client *client;

	TAILQ_FOREACH(client, &clientqueue, client_next) {
		if (!(client->flags & CLIENT_AUTHDONE))
			continue;

		if (memcmp(client->auth->client_id, id,
			   sizeof(client->auth->client_id)) == 0)
			break;
	}

	return (client);
}

/* Passed to callbacks */
struct client_run_arg {
	char jobid[128];
	int nr;
};

int
client_run_ready(int res, struct client *client, void *arg, char *line)
{
	struct client_run_arg *jobnr = arg;

	if (arg == NULL)
		return (IOFILT_DONE);

	if (res == -1) {
		struct schedule *sd;
		struct slot *slot;
		if ((sd  = schedule_find(jobnr->jobid)) == NULL)
			goto out;
		if ((slot = slot_find_nr(sd, jobnr->nr)) == NULL)
			goto out;
		schedule_returnslot(sd, slot->sl_slot);
		slot_free(slot);
	}
 out:
	free(arg);
	return (IOFILT_DONE);
}

/* Called when run preparation fails */

void
client_run_failed(struct client *client, struct schedule *sd, int nr)
{
	struct slot *slot = slot_find_nr(sd, nr);

	if (slot == NULL)
		return;

	schedule_returnslot(sd, slot->sl_slot);
	slot_free(slot);
}

int
client_run_prepare(struct schedule *sd, struct client *client)
{
	char line[1024], *arg;
	struct client_run_arg *jobnr;

	jobnr = malloc(sizeof(struct client_run_arg));
	if (jobnr == NULL)
		return (-1);
	strlcpy(jobnr->jobid, sd->sd_id, sizeof(jobnr->jobid));

	if (schedule_assign(sd, client, &jobnr->nr) == -1) {
		free(jobnr);
		return (-1);
	}
	arg = schedule_expand(sd, sd->sd_cmd, jobnr->nr);

	snprintf(line, sizeof(line), "%s %s %s", CMD_RUN, sd->sd_id, arg);

	/* Add this command to the comand queue */
	if (command_queue(&client->client_cmdqueue, line,
	    client_run_ready, jobnr) == -1) {
		/* Clean up */
		client_run_failed(client, sd, jobnr->nr);
		free(jobnr);

		return (-1);
	}

	return (0);
}

void
client_run(struct io_buffer *buf, struct schedule *sd)
{
	char line[128];
	struct client *client;
	int res;

	if (!client_onqueue(buf))
		return;

	TAILQ_FOREACH(client, &cmdqueue, cmd_next) {
		res = client_run_prepare(sd, client);

		if (res == -1) {
			snprintf(line, sizeof(line),
			    "Error: %s:%u failed to add cmd\n",
			    client->name, client->port);
			io_buffer_data(buf, line, strlen(line));
			return;
		}

		/* Just returns an error if a command is already running */
		command_start(client);
	}
}

void
client_run_cmd(struct io_buffer *buf, char *id, char *cmd)
{
	char line[1024], *arg;
	struct client *client;

	if (!client_onqueue(buf))
		return;

	TAILQ_FOREACH(client, &cmdqueue, cmd_next) {
		arg = schedule_expand(NULL, cmd, -1);
		snprintf(line, sizeof(line), "%s %s %s", CMD_RUN, id, arg);

		/* Add this command to the comand queue */
		if (command_queue(&client->client_cmdqueue, line,
			client_run_ready, NULL) == -1) {
			snprintf(line, sizeof(line),
			    "Error: %s:%u failed to add cmd '%s'\n",
			    client->name, client->port, cmd);
			io_buffer_data(buf, line, strlen(line));
			return;
		}

		/* Just returns an error if a command is already running */
		command_start(client);
	}
}

void
client_run_one(struct client *client, struct schedule *sd)
{
	if (client_run_prepare(sd, client) != -1)
		command_start(client);
}

#define MSG_NOCLIENT	"No clients connected.\n"
#define MSG_NOSELECTED	"No client selected for command execution.\n"

int
client_onqueue(struct io_buffer *buf)
{
	if (TAILQ_FIRST(&cmdqueue) == NULL) {
		io_buffer_data(buf,
			       MSG_NOSELECTED, sizeof(MSG_NOSELECTED) - 1);
		return (0);
	}

	return (1);
}

void
client_iterate(struct io_buffer *buf, char *regexp, int needrex, 
    int (*cb)(struct io_buffer *, struct client *, char *))
{
	char line[1024];
	struct client *tmp, *next;
	regex_t list;
	int i, userex = 0, negative = 0;

	if (TAILQ_FIRST(&clientqueue) == NULL) {
		io_buffer_data(buf, MSG_NOCLIENT, sizeof(MSG_NOCLIENT) - 1);
		return;
	}

	if (regexp != NULL && strlen(regexp)) {
		if (regexp[0] == '-') {
			negative = 1;
			regexp++;
		}

		if (regcomp(&list, regexp, REG_EXTENDED|REG_ICASE) == -1) {
			snprintf(line, sizeof(line),
			    "Invalid regular expression \"%s\"\n",
			    regexp);
			io_buffer_data(buf, line, strlen(line));
			return;
		}

		userex = 1;
	}

	if (needrex && !userex) {
		snprintf(line, sizeof(line),
		    "You need to specify a regular expression\n");
		io_buffer_data(buf, line, strlen(line));
		return;
	}

	client_now = time(NULL);
	i = 0;
	for (tmp = TAILQ_FIRST(&clientqueue); tmp; tmp = next) {
		next = TAILQ_NEXT(tmp, client_next);

		if (!(tmp->flags & CLIENT_AUTHDONE))
			continue;

		i++;
		client_statusline(line, sizeof(line), i, tmp);
		if (userex) {
			if (negative ^ (regexec(&list, line, 0, NULL, 0) != 0))
				continue;
		}
		if ((*cb)(buf, tmp, line) == -1)
			break;
	}

	if (userex)
		regfree(&list);
}

int
client_list_cb(struct io_buffer *buf, struct client *client, char *line)
{
	return (io_buffer_data(buf, line, strlen(line)));
}

void
client_list(struct io_buffer *buf, char *regexp)
{
	client_iterate(buf, regexp, 0, client_list_cb);
}

int
client_kill_cb(struct io_buffer *buf, struct client *client, char *line)
{
	io_terminate((struct io_obj *)client->src, IORES_END);

	return (0);
}

void
client_kill(struct io_buffer *buf, char *regexp)
{
	client_iterate(buf, regexp, 1, client_kill_cb);
}

int
client_select(struct io_buffer *buf, char *regexp)
{
	struct client *client;
	char line[1024];
	int n, i, negative = 0;
	regex_t select;

	if (regexp[0] == '-') {
		negative = 1;
		regexp++;
	}

	if (regcomp(&select, regexp, REG_EXTENDED|REG_ICASE) == -1) {
		snprintf(line, sizeof(line),
			 "Invalid regular expression \"%s\"\n",
			 regexp);
		io_buffer_data(buf, line, strlen(line));
		return (-1);
	}

	client_now = time(NULL);
	n = 0;
	i = 0;
	TAILQ_FOREACH(client, &clientqueue, client_next) {
		if (!(client->flags & CLIENT_AUTHDONE))
			continue;

		i++;

		if (negative) {
			if (!(client->flags & CLIENT_ONCMDLIST))
				continue;
		} else {
			if (client->flags & CLIENT_ONCMDLIST)
				continue;
		}

		client_statusline(line, sizeof(line), i, client);
		if (regexec(&select, line, 0, NULL, 0) == 0) {
			/* Put client on cmd list */
			if (negative) {
				client->flags &= ~CLIENT_ONCMDLIST;
				TAILQ_REMOVE(&cmdqueue, client, cmd_next);
				client_selected--;
			} else {
				client->flags |= CLIENT_ONCMDLIST;
				TAILQ_INSERT_TAIL(&cmdqueue, client, cmd_next);
				client_selected++;
			}
			n++;
		}
	}

	snprintf(line, sizeof(line), "%s %d clients.\n",
		 negative ? "Deselected" : "Selected", n);
	io_buffer_data(buf, line, strlen(line));

	regfree(&select);

	return (0);
}

#define MSG_NOCLISEL	"No client selected for execution.\n"

void
client_parsestatus(char *data, size_t size, struct client *client)
{
	char tmp[128], *p, *token;
	
	data += CMD_SZ(CMD_STATUS);
	data += strspn(data, " :");

	/* Check if this is a new status */
	if (strcmp(client->status, data) == 0)
		return;
	
	strlcpy(client->status, data, sizeof(client->status));

	/* Do extra parsing for orphaned jobs */
	strlcpy(tmp, data, sizeof(tmp));
	p = tmp;

	while ((token = strnsep(&p, ", ")) != NULL) {
		char *job, *pid;

		job = strnsep(&token, "(");
		pid = strnsep(&token, ")");

		/* See if we found job and pid */
		if (job == NULL || strlen(job) == 0 ||
		    pid == NULL || strlen(pid) == 0)
			continue;

		/* Reconnect only if the job has a PID */
		if (atoi(pid) == 0)
			continue;

		/* See if we can reconnect this job */
		schedule_clientconnect(client, job);
	}
}

void
client_parseinfo(char *data, size_t size, struct client *client)
{
	char jobid[128];
	int reason;

	data += CMD_SZ(CMD_INFO);
	data += strspn(data, " :");

	jobid[127] = '\0';
	if (sscanf(data, "%d %127s", &reason, jobid) != 2)
		return;

	fprintf(stderr, __FUNCTION__": %s:%u -> %s %s\n",
		client->name, client->port, jobid,
		reason == INFO_CHLDDEAD ? "died" : "finished");

	schedule_slotdone(client, jobid, reason == INFO_CHLDDEAD);
}

void
client_parseoutput(char *data, size_t size, struct client *client)
{
	char jobid[128], line[1024];
	struct schedule *sd;
	struct slot *slot;
	char which, *end, *start;
	int slotnr = -1, outfd;
	time_t now;
	size_t jobsize;

	/* Remove terminating 0 */
	size--;

	start = data;
	data += CMD_SZ(CMD_OUTPUT);
	data += strspn(data, " :");

	/* Save channel and parameters */
	which = *data;
	end = strchr(data + 2, ':');
	if (data[1] != ' ' || (which != 'O' && which != 'E') ||
	    end == NULL) {
	    baddata:
		fprintf(stderr, __FUNCTION__": malformed output from %s:%u\n",
			client->name, client->port);
		return;
	}
	data += 2;

	jobsize = end - data;
	if (jobsize == 0 || jobsize > sizeof(jobid) - 1)
		goto baddata;

	memcpy(jobid, data, jobsize);
	jobid[jobsize] = '\0';

	sd = schedule_find(jobid);
	if (sd == NULL) {
		fprintf(stderr, __FUNCTION__
			": output from %s:%u for unknown id %s\n",
			client->name, client->port, jobid);
		return;
	}
	slot = slot_find_id(sd, client->auth->client_id);
	if (slot != NULL)
		slotnr = slot->sl_slot;

	size -= end + 2 - start;
	data = end + 2;
	outfd = which == 'O' ? sd->sd_stdout : sd->sd_stderr;

        now = time(NULL);
	snprintf(line, sizeof(line), "%.24s: %s:%u:%d:",
	    ctime(&now), client->name, client->port, slotnr);

	/* Check if it's binary data */
	if (strlen(data) == size) {
		while ((end = strchr(data, '\n')) != NULL) {
			write(outfd, line, strlen(line));
			write(outfd, data, end - data + 1);

			/* Advance */
			size -= end + 1 - data;
			data = end + 1;
		}
	}

	if (size) {
		/* Prepend from whom the information is coming */
		write(outfd, line, strlen(line));
		write(outfd, data, size);
	}

	return;
}

/* This parses the client responses */

int
filter_clientresponse(void *arg, struct io_filter *filter,
		      struct io_buffer *buf, int flags)
{
 	struct client *client = arg;
	struct io_buffer *mybuf = &filter->io_buf;
	char line[1024];
	int res, cbres;
	void *data;
	size_t size;

	if (io_buffer_append(mybuf, buf) == -1)
		return (IOFILT_ERROR);

	for (; (res = io_buffer_haspacket(mybuf, &data, &size)) == IOFILT_DONE;
	     io_buffer_drainpacket(mybuf)) {
		if (((u_char *)data)[size - 1] != '\0') {
			snprintf(line, sizeof(line), __FUNCTION__
				 ": illegal client response: %s:%u\n",
				 client->name, client->port);
			ui_infocb(client, line);
			continue;
		}

		if (timeout_pending(&client->ev, NULL))
			client_settimeout(client);

		if (!strncasecmp(data, CMD_STATUS, CMD_SZ(CMD_STATUS))) {
			client_parsestatus(data, size, client);
			continue;
		} else if (!strncasecmp(data, CMD_INFO, CMD_SZ(CMD_INFO))) {
			client_parseinfo(data, size, client);

			continue;
		} else if (!strncasecmp(data, CMD_OUTPUT,
					CMD_SZ(CMD_OUTPUT))) {
			client_parseoutput(data, size, client);
			continue;
		}


		/* XXX - done */
		if (!(client->flags & CLIENT_CMDINPROGRESS)) {
			snprintf(line, sizeof(line), __FUNCTION__
				 ": got cmd response but not waiting: %s:%u\n",
				 client->name, client->port);
			io_buffer_data(&client_errbuf, line, strlen(line));
			if (!client->tap)
				continue;
			ui_infocb(client, line);
		}

		cbres = 0;
		/* Examine different responses from client */
		if (!strncasecmp(data, CMD_ERROR, CMD_SZ(CMD_ERROR))) {
			snprintf(line, sizeof(line), "%s:%u cmd: %s\n",
			    client->name, client->port, (char *)data);
			io_buffer_data(&client_errbuf, line, strlen(line));
			if (client->tap)
				ui_infocb(client, line);
			client_errors++;
			cbres = -1;
			/* FALL THROUGH */
		} else if (!strncasecmp(data, CMD_DONE, CMD_SZ(CMD_DONE))) {
			if (client->tap) {
				snprintf(line, sizeof(line), "%s:%u cmd: %s\n",
				    client->name, client->port, (char *)data);
				ui_infocb(client, line);
			}
			client_done++;
			cbres = 1;
			/* FALL THROUGH */
		}

		if (command_reply(client, data, cbres) == IOFILT_DONE) {
			/* Start another command for this client */
			command_start(client);
		}
	}

	return (res);
}
