/*
 * Copyright 2001 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/queue.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <unistd.h>
#include <string.h>
#include <libgen.h>

#include "config.h"

#include <event.h>
#include <io.h>

#include "server.h"
#include "io_aux.h"
#include "util.h"
#include "commands.h"
#include "schedule.h"
#include "ui.h"

char *ui_file = UI_FIFO;

extern struct io_filter *anchor;
extern struct clientlist cmdqueue;

struct client_serialize_queue *csq_init;
struct io_duplex *ui_dplx = NULL;

char cmd_jobid[128];

#define PROMPT	"discern> "

struct command {
	char *cmd;
	char *short_help;
	char *long_help;
	int (*func)(struct io_buffer *, char *);
};

int ui_write_prompt(struct io_obj *);
int ui_buffer_prompt(struct io_buffer *);

int ui_command_help(struct io_buffer *, char *);
int ui_command_list(struct io_buffer *, char *);
int ui_command_kill(struct io_buffer *, char *);
int ui_command_select(struct io_buffer *, char *);
int ui_command_init(struct io_buffer *, char *);
int ui_command_job(struct io_buffer *, char *);
int ui_command_jobs(struct io_buffer *, char *);
int ui_command_run(struct io_buffer *, char *);
int ui_command_query(struct io_buffer *, char *);
int ui_command_show(struct io_buffer *, char *);
int ui_command_drain(struct io_buffer *, char *);
int ui_command_reclaim(struct io_buffer *buf, char *line);

struct command commands[] = {
	{
		"job",
		"job\t\t sets up parameters for a distributed job\n",
		"Configures a job: job <jobid> [...]\n"
		"\t [-del|-new]\t delete/create new job\n",
		ui_command_job
	},
	{
		"jobs",
		"jobs\t\t lists all currently configured jobs\n",
		"Lists all currently configured jobs.\n",
		ui_command_jobs
	},
	{
		"init",
		"init\t\t prepares selected clients for job\n",
		"init <jobid> <dataset>\n",
		ui_command_init
	},
	{
		"run",
		"run\t\t starts remote execution of the configured job\n",
		"Executes a configured job:\n"
		"run jobid\n",
		ui_command_run
	},
	{
		"show",
		"show\t\t shows buffered messages\n",
		"show <errors|messages>\n",
		ui_command_show
	},
	{
		"drain",
		"drain\t\t drain buffered messages\n", 
		"drain <errors|messages|all>\n",
		ui_command_drain
	},
	{
		"query",
		"query\t\t queries the state of a dataset\n",
		"query <jobid> <filename>\n",
		ui_command_query
	},
	{
		"select",
		"select\t\t [de]selects clients for command execution\n",
		"select [-]<regexp>\n",
		ui_command_select
	},
	{
		"list",
		"list\t\t lists currently connected clients\n",
		"list [-]<regexp>\n",
		ui_command_list
	},
	{
		"kill",
		"kill\t\t disconnects clients\n",
		"kill <regexp>\n",
		ui_command_kill
	},
	{
		"reclaim",
		"reclaim\t\t reclaim orphaned jobs\n",
		"reclaim <jobid>\n",
		ui_command_reclaim
	},
	{
		"help",
		"help\t\t outputs a command help\n",
		"help [command]\n",
		ui_command_help
	},
	{
		NULL, NULL, NULL, NULL
	}
};

char tmpbuf[1024];

char *
make_prompt(void)
{
	extern int client_errors, client_done;
	extern int client_connected, client_selected;
	extern struct io_buffer client_errbuf, client_donebuf;
	static char tmp[128];
	char wait[32], csq[32];

	wait[0] = '\0';
	csq[0] = '\0';

	/* Show waiting messages */
	if (client_errbuf.off || client_donebuf.off)
		snprintf(wait, sizeof(wait), ":%s%s",
		    client_errbuf.off ? "e" : "",
		    client_donebuf.off ? "m" : "");
	if (csq_init->csq_current)
		snprintf(csq, sizeof(csq), ":F%d/%d[%d]",
		    csq_init->csq_current, csq_init->csq_max,
		    csq_init->csq_pending);
		    
	snprintf(tmp, sizeof(tmp), "%dE %dD %d/%dC%s%s:%s",
	    client_errors, client_done,
	    client_selected, client_connected, wait, csq,
	    PROMPT);

	client_errors = client_done = 0;

	return (tmp);
}

int
ui_write_prompt(struct io_obj *sink)
{
	u_char *tmp = make_prompt();
	return (io_data_write(sink, tmp, strlen(tmp)));
}

int
ui_buffer_prompt(struct io_buffer *buf)
{
	u_char *tmp = make_prompt();
	return (io_buffer_data(buf, tmp, strlen(tmp)));
}

void
ui_dead(struct io_obj *obj, void *arg)
{
	ui_dplx = NULL;

	fprintf(stderr, __FUNCTION__": ui on fd %d is gone\n",
		obj->io_ev.ev_fd);
}

void
ui_infocb(struct client *client, char *msg)
{
	if (ui_dplx == NULL)
		return;
	if (msg[0] != '\n')
		io_data_write(ui_dplx->io_dst, "\n", 1);
	io_data_write(ui_dplx->io_dst, msg, strlen(msg));
	ui_write_prompt(ui_dplx->io_dst);
}

void
ui_error_cb(struct client *client, char *msg)
{
	ui_infocb(client, msg);
}

int
ui_command_drain(struct io_buffer *buf, char *line)
{
	extern struct io_buffer client_errbuf, client_donebuf;
	char *what;

	what = strnsep(&line, WHITESPACE);
	if (what == NULL || strlen(what) == 0)
		return (-1);

	if (strcmp(what, "errors") == 0) {
		io_buffer_drain(&client_errbuf, client_errbuf.off);
	} else if (strcmp(what, "messages") == 0) {
		io_buffer_drain(&client_donebuf, client_donebuf.off);
	} else if (strcmp(what, "all") == 0) {
		io_buffer_drain(&client_errbuf, client_errbuf.off);
		io_buffer_drain(&client_donebuf, client_donebuf.off);
	} else
		return (-1);

	return (0);
}

int
ui_command_show(struct io_buffer *buf, char *line)
{
	extern struct io_buffer client_errbuf, client_donebuf;
	char *what;

	what = strnsep(&line, WHITESPACE);
	if (what == NULL || strlen(what) == 0)
		return (-1);

	if (strcmp(what, "errors") == 0) {
		client_errbuf.ready = client_errbuf.off;
		io_buffer_append(buf, &client_errbuf);
		io_buffer_drain(&client_errbuf, client_errbuf.off);
	} else if (strcmp(what, "messages") == 0) {
		client_donebuf.ready = client_donebuf.off;
		io_buffer_append(buf, &client_donebuf);
		io_buffer_drain(&client_donebuf, client_donebuf.off);
	} else
		return (-1);

	return (0);
}

int
ui_command_run(struct io_buffer *buf, char *line)
{
	char *jobid;
	struct schedule *sd;

	jobid = strnsep(&line, WHITESPACE);
	if (jobid == NULL || strlen(jobid) == 0)
		return (-1);

	sd = schedule_find(jobid);
	if (sd == NULL) {
		snprintf(tmpbuf, sizeof(tmpbuf),
		    "Error: unknown job id \"%s\"\n",
		    jobid);
		io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
		return (0);
	}

	if (line == NULL || strlen(line) == 0) {
		if (sd->sd_cmd == NULL) {
			snprintf(tmpbuf, sizeof(tmpbuf),
			    "Error: job id \"%s\" not set up.\n",
			    jobid);
			io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
			return (0);
		}

		client_run(buf, sd);
	} else
		client_run_cmd(buf, jobid, line);

	return (0);
}

int ui_query_ready(int, struct client *, void *, char *);

void
ui_query_current(struct client *client, struct dataset *ds)
{
	char tmp[256];
	struct dataset *tmpsf;

	if ((tmpsf = malloc(sizeof(struct dataset))) == NULL) {
		warn(__FUNCTION__": command aborted: malloc");
		return;
	}
	memcpy(tmpsf, ds, sizeof(struct dataset));

	snprintf(tmp, sizeof(tmp), "%s %s %s",
		 CMD_QUERY, cmd_jobid, basename(ds->sf_name));

	if (command_queue(&client->client_cmdqueue, tmp,
		ui_query_ready, tmpsf) == -1) {
		free(tmpsf);
		return;
	}

	command_start(client);
}

int
ui_query_ready(int res, struct client *client, void *arg, char *line)
{
	extern struct io_buffer client_donebuf;
	char *sf_name, *sf_size, *hash, *p;
	struct dataset tmpsf, *sf = arg;

	if (res == -1) {
		free(sf);
		return (IOFILT_ERROR);
	}
	res = IOFILT_ERROR;

	strlcpy(tmpbuf, line, sizeof(tmpbuf));
	p = tmpbuf;
	strnsep(&p, ": ");
	sf_name = strnsep(&p, " ");
	sf_size = strnsep(&p, " ");
	hash = strnsep(&p, " ");

	if (sf_name == NULL || sf_size == NULL || hash == NULL ||
	    strlen(hash) != sizeof(tmpsf.sf_hash) * 2) {
		snprintf(tmpbuf, sizeof(tmpbuf),
		    "Error: %s:%u: Query response invalid: %s\n",
		    client->name, client->port, line);
		goto out;
	}

	memset(&tmpsf, 0, sizeof(struct dataset));
	strlcpy(tmpsf.sf_name, sf_name, sizeof(tmpsf.sf_name));
	/* XXX - quad_t */
	tmpsf.sf_size = strtoul(sf_size, &p, 0);
	hex2bin(tmpsf.sf_hash, hash, sizeof(tmpsf.sf_hash));

	res = dataset_verify(sf, &tmpsf) ? IOFILT_DONE : IOFILT_ERROR; 
	snprintf(tmpbuf, sizeof(tmpbuf), "Info: %s:%u: %s verification %s\n",
	    client->name, client->port, sf->sf_name, 
	    res == IOFILT_DONE ? "OK" : "ERROR");
 out:
	io_buffer_data(&client_donebuf, tmpbuf, strlen(tmpbuf));
	if (client->tap)
		ui_infocb(client, tmpbuf);

	free(sf);
	return (res);
}

int
ui_command_query(struct io_buffer *buf, char *line)
{
	struct client *client;
	struct client_cmd *script_start;

	char *jobid;
	struct schedule *sd;
	int i;

	jobid = strnsep(&line, WHITESPACE);
	if (jobid == NULL || strlen(jobid) == 0)
		return (-1);

	strlcpy(cmd_jobid, jobid, sizeof(cmd_jobid));

	sd = schedule_find(jobid);
	if (sd == NULL) {
		snprintf(tmpbuf, sizeof(tmpbuf),
			 "Error: unknown job id \"%s\"\n",
			 jobid);
		io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
		return (0);
	}

	if (!sd->sd_nfiles) {
		snprintf(tmpbuf, sizeof(tmpbuf),
			 "Error: no files for job id \"%s\"\n",
			 jobid);
		io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
		return (0);
	}

	if (!client_onqueue(buf))
		return (0);

	TAILQ_FOREACH(client, &cmdqueue, cmd_next) {
		script_start = command_script_start(&client->client_cmdqueue);
		for (i = 0; i < sd->sd_nfiles; i++)
			ui_query_current(client, &sd->sd_files[i]);
		command_script_end(&client->client_cmdqueue, script_start);
	}

	return (0);
}

int
ui_command_jobs(struct io_buffer *buf, char *line)
{
	schedule_list(buf);

	return (0);
}


int
ui_command_reclaim(struct io_buffer *buf, char *line)
{
	char *jobid;
	struct schedule *sd;

	jobid = strnsep(&line, WHITESPACE);
	if (jobid == NULL || strlen(jobid) == 0)
		return (-1);

	sd = schedule_find(jobid);
	schedule_reclaim(sd, buf);

	return (0);
}

int
ui_command_job(struct io_buffer *buf, char *line)
{
	char *jobid, *p, *tok, *arg;
	struct schedule *sd;
	int maymodify = 1, verbose = 0;

	jobid = strnsep(&line, WHITESPACE);
	if (jobid == NULL || strlen(jobid) == 0)
		return (-1);

	sd = schedule_find(jobid);
	if (sd == NULL) {
		tok = strnsep(&line, WHITESPACE);
		if (tok == NULL || strcasecmp(tok, "-new")) {
			snprintf(tmpbuf, sizeof(tmpbuf),
				 "Error: unknown job id \"%s\"\n",
				 jobid);
			io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
			return (0);
		}

		if (!file_checkname(jobid)) {
			snprintf(tmpbuf, sizeof(tmpbuf),
				 "Error: illegal job id \"%s\"\n",
				 jobid);
			io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
			return (0);
		}
		
		sd = schedule_new(jobid);
		if (sd == NULL)
			goto error;
	} else
		if (line != NULL && !strncasecmp(line, "-del", 4)) {
			if (sd->sd_flags & SDFLAGS_LOCKED) {
				snprintf(tmpbuf, sizeof(tmpbuf),
					 "Warning: job id \"%s\" is locked\n",
					 jobid);
				io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
				return (0);
			}
			schedule_free(sd);
			return (0);
		}

	/* Check if we assigned slots already, then this struct is fixed */
	if (line != NULL && strlen(line) != 0 &&
	    SD_HAS_RANGE(sd) && sd->sd_left != sd->sd_nbits) {
		snprintf(tmpbuf, sizeof(tmpbuf),
			 "Warning: job id \"%s\" is running, some modifications impossible\n",
			 jobid);
		io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
		maymodify = 0;
	}

	while ((tok = strnsep(&line, WHITESPACE)) != NULL) {
		if (!strcasecmp(tok, "-new") && maymodify) {
			snprintf(tmpbuf, sizeof(tmpbuf),
				 "Warning: ignoring -new\n");
			io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
		} else if (!strcasecmp(tok, "-cmd") && maymodify) {
			if (line == NULL || strlen(line) == 0)
				return (-1);

			if (sd->sd_cmd != NULL)
				free(sd->sd_cmd);

			if ((sd->sd_cmd = strdup(line)) == NULL)
				goto error;
			break;
		} else if (!strcasecmp(tok, "-range") && maymodify) {
			arg = strnsep(&line, WHITESPACE);
			if (arg == NULL || strlen(arg) == NULL)
				return (-1);

			if (sscanf(arg, "%d-%d",
				   &sd->sd_start, &sd->sd_end) != 2)
				return (-1);

			schedule_update(sd);
		} else if(!strcasecmp(tok, "-automatic")) {
			sd->sd_flags ^= SDFLAGS_AUTOMATIC;
		} else if(!strcmp(tok, "-v"))
			verbose = 1;
	}

	p = schedule_status(sd, verbose);
	snprintf(tmpbuf, sizeof(tmpbuf), "%s", p);
	io_buffer_data(buf, tmpbuf, strlen(tmpbuf));

	return (0);

 error:
	snprintf(tmpbuf, sizeof(tmpbuf),
		 "Error: %s: %s\n",
		 jobid, strerror(errno));
	io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
	return (0);
}

/* Call backs for the init command */

struct ui_init_arg {
	char filename[255];
	int fd;
};

void
ui_init_done(struct io_obj *obj, void *arg)
{
	struct ui_init_arg *uia = arg;
	fprintf(stderr, __FUNCTION__": fd: %d, %s: reading done\n",
	    uia->fd, uia->filename);
}

void
ui_abort(struct client *client)
{
	io_filter_packet(client->dst, CMD_ABORT, sizeof(CMD_ABORT));
}

int
ui_init_ready(int res, struct client *client, void *arg, char *line)
{
	struct ui_init_arg *uia = arg;
	struct io_obj *src;
	extern struct io_buffer client_errbuf, client_donebuf;

	/* How to deal with errors in this stage? */
	if (client->flags & CLIENT_CMDSTAGED) {
		client->flags &= ~CLIENT_CMDSTAGED;
		
		snprintf(tmpbuf, sizeof(tmpbuf),
			 "Info: %s:%u finished file transfer: %s\n",
			 client->name, client->port, uia->filename);
		io_buffer_data(&client_donebuf, tmpbuf, strlen(tmpbuf));
		if (client->tap)
			ui_infocb(client, tmpbuf);
		free(uia);
		return (IOFILT_DONE);
	}

	if (res == -1)
		goto cleanup;

	if ((uia->fd = open(uia->filename, O_RDONLY, 0)) == -1 ||
	    (src = io_new_source(uia->fd)) == NULL) {
		snprintf(tmpbuf, sizeof(tmpbuf),
		    "Error: %s:%u file %s(%d) transfer error\n",
		    client->name, client->port, uia->filename, uia->fd);
		io_buffer_data(&client_errbuf, tmpbuf, strlen(tmpbuf));
		if (client->tap)
			ui_infocb(client, tmpbuf);

		ui_abort(client);

		goto cleanup;
	}

	src->io_free = ui_init_done;
	src->io_freearg = uia;

	io_attach(src, client->dst);

	/* Mark so that we know about the second stage of the cmd */
	client->flags |= CLIENT_CMDSTAGED;

	fprintf(stderr, __FUNCTION__": %s:%u: started file transfer\n",
		client->name, client->port);

	return (IOFILT_NEEDMORE);

cleanup:
	if (uia->fd != -1)
		close(uia->fd);
	free(uia);
	return (IOFILT_DONE);
}

int
ui_command_init(struct io_buffer *buf, char *line)
{
	struct client *client;
	char *jobid, *file;
	struct schedule *sd;
	struct dataset *sf;
	struct ui_init_arg *uia;

	if (line == NULL)
		return (-1);

	line += strspn(line, WHITESPACE);

	jobid = strnsep(&line, WHITESPACE);
	if (line == NULL || *line == '\0')
		return (-1);
	file = strnsep(&line, WHITESPACE);
	if (line != NULL)
		return (-1);

	if ((sd = schedule_find(jobid)) == NULL &&
	    (sd = schedule_new(jobid)) == NULL) {
		snprintf(tmpbuf, sizeof(tmpbuf), "Error: can not create %s\n",
			 jobid);
		io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
		return (0);
	}

	if ((sf = dataset_findfile(sd->sd_files, &sd->sd_nfiles,
				   NULL, file)) == NULL &&
	    (sf = dataset_addfile(sd->sd_files, &sd->sd_nfiles,
				  file, NULL)) == NULL) {
		snprintf(tmpbuf, sizeof(tmpbuf),
			 "Error: schedule_addfile(\"%s\"): %s\n",
			 file, strerror(errno));
		io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
		return (0);
	}

	if (!client_onqueue(buf))
		return (0);

	snprintf(tmpbuf, sizeof(tmpbuf), "%s %s %s %lld",
		 CMD_INIT, jobid, basename(file), sf->sf_size);

	TAILQ_FOREACH(client, &cmdqueue, cmd_next) {
		uia = malloc(sizeof(struct ui_init_arg));
		if (uia == NULL) {
			snprintf(tmpbuf, sizeof(tmpbuf),
				 "Error: malloc(%d): %s\n",
				 sizeof(struct ui_init_arg), strerror(errno));
			io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
			return (0);
		}
		strlcpy(uia->filename, file, sizeof(uia->filename));
		uia->fd = -1;

		if (command_queue(&client->client_cmdqueue, tmpbuf,
				  ui_init_ready, uia) == -1) {
			free(uia);
			snprintf(tmpbuf, sizeof(tmpbuf),
				 "Error: open(\"%s\"): %s\n",
				 file, strerror(errno));
			io_buffer_data(buf, tmpbuf, strlen(tmpbuf));
			return (0);
		}
		command_serialize_last(csq_init, client);
		command_start(client);
	}

	return (0);
}

int
ui_command_select(struct io_buffer *buf, char *line)
{
	if (line == NULL)
		return (-1);

	client_select(buf, line);

	return (0);
}

int
ui_command_list(struct io_buffer *buf, char *line)
{
	client_list(buf, line);

	return (0);
}

int
ui_command_kill(struct io_buffer *buf, char *line)
{
	client_kill(buf, line);

	return (0);
}

int
ui_command_help(struct io_buffer *buf, char *line)
{
	char output[1024];
	struct command *cmd;
	char *command;

	command = strnsep(&line, WHITESPACE);
	if (command != NULL && strlen(command)) {
		for (cmd = commands; cmd->cmd; cmd++) {
			/* Find out what command was sent.  */
			if (strcasecmp(cmd->cmd, command) == 0)
				break;
		}
		if (cmd->cmd == NULL) {
			snprintf(output, sizeof(output),
			    "Error: unknown command \"%s\"\n", command);
			io_buffer_data(buf, output, strlen(output));
			return (0);
		}
		io_buffer_data(buf, cmd->long_help, strlen(cmd->long_help));
	
	} else {
		for (cmd = commands; cmd->cmd; cmd++)
			io_buffer_data(buf,
			    cmd->short_help, strlen(cmd->short_help));
	}

	return (0);
}

void
ui_handle_command(struct io_buffer *buf, char *line)
{
	char output[1024];
	char *command;
	struct command *cmd;

	command = strnsep(&line, WHITESPACE);
	if (!strlen(command))
		return;

	for (cmd = commands; cmd->cmd; cmd++) {
		/* Find out what command was sent.  */
		if (strcasecmp(cmd->cmd, command) == 0)
			break;
	}
	
	if (!cmd->cmd) {
		snprintf (output, sizeof(output),
			 "ui_handle_command: unrecognized command: '%s'\n",
			 command);
		io_buffer_data(buf, output, strlen(output));
		return;
	}

	if ((*cmd->func)(buf, line) == -1) {
		snprintf(output, sizeof(output), "%s%s",
			 "ui_handle_command: missing arguments\n",
			 cmd->short_help);
		io_buffer_data(buf, output, strlen(output));
	}

	return;
}

int
ui_handler(void *arg, struct io_filter *filter, struct io_buffer *buf,
	   int flags)
{
	struct io_buffer *mybuf = &filter->io_tmpbuf;
	struct io_buffer *outbuf = &filter->io_buf;
	char *p;
	int n;

	if (io_buffer_append(mybuf, buf) == -1)
		return (IOFILT_ERROR);

	n = mybuf->off;
	p = mybuf->data;
	while (n--) {
		/*
		 * When we find a newline, cut off the line and feed it to the
		 * command processor.  Then move the rest up-front.
		 */
		if (*p == '\n') {
			*p = '\0';
			ui_handle_command(outbuf, mybuf->data);

			memmove(mybuf->data, p + 1, n);
			p = mybuf->data;
			mybuf->off = n;
			continue;
		}
		p++;
	}

	if (ui_buffer_prompt(outbuf) == IOFILT_ERROR)
		return (IOFILT_ERROR);

	outbuf->ready = outbuf->off;

	return (IOFILT_DONE);
}

int
ui_new(void *arg, struct io_filter *filter, struct io_buffer *buf, int flags)
{
	struct io_buffer *mybuf = &filter->io_buf;
	struct io_filter *filt;
	int newfd;

	if (buf->ready < sizeof (int))
		return (IOFILT_NEEDMORE);

	if (io_buffer_extend(mybuf, 256) == -1)
		return (IOFILT_ERROR);

	newfd = *(int *)buf->data;

	/* Close old connection */
	if (ui_dplx != NULL)
		io_terminate(ui_dplx->io_src, IORES_END);

	if ((ui_dplx = io_new_socket(newfd)) == NULL)
		return (IOFILT_NEEDMORE);

	/* Remember when we die */
	ui_dplx->io_src->io_free = ui_dead;
	ui_dplx->io_src->io_freearg = NULL;

	fprintf(stderr, __FUNCTION__": New ui connection on fd %d\n", newfd);

	filt = io_new_filter(ui_handler, NULL);
	io_attach(ui_dplx->io_src, filt);
	io_attach(filt, ui_dplx->io_dst);

	ui_write_prompt(ui_dplx->io_dst);

	return (IOFILT_NEEDMORE);
}

void
ui_init(void)
{
        struct stat st;
        struct sockaddr_un ifsun;
	struct io_obj *fdsrc;
	struct io_filter *filt;
	int ui_socket;

	/* Create serialize queue for init */
	if ((csq_init = command_serialize_new(CSQ_INIT_MAX)) == NULL)
		err(1, __FUNCTION__": command_serialize_new()");

        /* Don't overwrite a file */
        if (lstat(ui_file, &st) == 0)
                if ((st.st_mode & S_IFMT) == S_IFREG) {
                        errno = EEXIST;
                        err(1, __FUNCTION__": could not create FIFO: %s",
			    ui_file);
                }

        /* No need to know about errors.  */
        unlink(ui_file);

        ui_socket = socket(AF_UNIX, SOCK_STREAM, 0);
        if (ui_socket == -1)
                err(1, __FUNCTION__": socket");
        if (setsockopt(ui_socket, SOL_SOCKET, SO_REUSEADDR,
                       &ui_socket, sizeof (ui_socket)) == -1)
                err(1, __FUNCTION__": setsockopt");

        memset(&ifsun, 0, sizeof (ifsun));
        ifsun.sun_family = AF_UNIX;
        strlcpy(ifsun.sun_path, ui_file, sizeof(ifsun.sun_path));
#ifdef HAVE_SUN_LEN
        ifsun.sun_len = strlen(ifsun.sun_path);
#endif /* HAVE_SUN_LEN */
        if (bind(ui_socket, (struct sockaddr *)&ifsun, sizeof (ifsun)) == -1)
                err(1, __FUNCTION__": bind");

        if (listen(ui_socket, 5) == -1)
                err(1, __FUNCTION__": listen");

	fdsrc = io_new_obj(ui_socket, IOTYPE_SOURCE,
			   io_method_accept, IO_ACCEPTSIZE);
	if (fdsrc == NULL)
		errx(1, __FUNCTION__": io_new_obj");

	filt = io_new_filter(ui_new, NULL);
	io_attach(fdsrc, filt);
}
