/*
 * Copyright 2000 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/uio.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>
#include <libgen.h>

#include "util.h"

char h2btab[] = {
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
	'a', 'b', 'c', 'd', 'e', 'f'
};

#define nibbletobin(y)	((y) >= 'a' ? 10 + (y) - 'a' : (y) - '0')
#define hextobin(x)	(nibbletobin((x)[0])*16 + nibbletobin((x)[1]))

int
hex2bin(u_char *bin, char *hex, size_t len)
{
	char asc[2];
	int i;

	if (strlen(hex) / 2 < len)
		return (-1);

	for (i = 0; i < len; i++) {
		asc[0] = tolower(hex[2*i]);
		asc[1] = tolower(hex[2*i + 1]);	
		if (!isxdigit(asc[0]) || !isxdigit(asc[1]))
			return (-1);
		bin[i] = hextobin(asc);
	}

	return (i);
}

void
bin2hex(char *hex, u_char *bin, size_t len)
{
	int i;

	for (i = 0; i < len; i++) {
		hex[i*2 + 0] = h2btab[bin[i] >> 4];
		hex[i*2 + 1] = h2btab[bin[i] & 0xf];
	}
	hex[i*2 + 0] = '\0';
}

u_char *
memfind(u_char *data, size_t len, u_char *find, size_t flen)
{
	char *where;
	size_t off;

	while (len > 0) {
		where = memchr(data, find[0], len);
		if (where == NULL)
			return (NULL);
		off = (size_t)where - (size_t)data;
		if (len - off < flen)
			return (NULL);
		if (memcmp(where, find, flen) == 0)
			return (where);
		else
			off += 2;

		data += off;
		len -= off;
	}

	return (NULL);
}

char *
strnsep(char **line, char *delim)
{
	char *ret, *p;

	if (line == NULL)
		return (NULL);

	ret = *line;
	if (ret == NULL)
		return (NULL);

	p = strpbrk(ret, delim);
	if (p == NULL) {
		*line = NULL;
		return (ret);
	}

	*line = p + strspn(p, delim);
	*p = '\0';

	return (ret);
}

int
file_hasextension(char *name, char *ext)
{
	int nlen = strlen(name);
	int elen = strlen(ext);

	if (nlen < elen)
		return (0);

	return (strcasecmp(name + nlen - elen, ext) == 0);
}


char *
strrpl(char *str, size_t size, char *match, int value)
{
	char tmp[20];
	char *p, *e;
	int len, rlen;

	p = str;
	e = p + strlen(p);
	len = strlen(match);

	/* Try to match against the variable */
	while ((p = strchr(p, match[0])) != NULL) {
		if (!strncmp(p, match, len) && !isalnum(p[len]))
			break;
		p += len;

		if (p >= e)
			return (NULL);
		    
	}

	if (p == NULL)
		return (NULL);

	snprintf(tmp, sizeof(tmp), "%d", value);
	rlen = strlen(tmp);

	if (strlen(str) - len + rlen > size)
		return (NULL);

	memmove(p + rlen, p + len, strlen(p + len) + 1);
	memcpy(p, tmp, rlen);

	return (p);
}

char *
elapsedtime(time_t duration)
{
	static char tmp[128];
	int minutes = duration / 60;
	int hours = minutes / 60;
	int days = hours / 24;

	if (!minutes) {
		snprintf(tmp, sizeof(tmp), "%d sec", duration);
	} else if (!hours) {
		snprintf(tmp, sizeof(tmp), "%d:%.2d min",
			 minutes, duration % 60);
	} else if (!days) {
		snprintf(tmp, sizeof(tmp), "%d:%.2d hr",
			 hours, minutes % 60);
	} else {
		snprintf(tmp, sizeof(tmp), "%d days %d hr", days, hours % 24);
	}

	return (tmp);
}

char *
time2rate(float duration)
{
	static char tmp[128];
	int multiply;
	char *rate;

	if (duration < 2) {
		multiply = 1;
		rate = "/sec";
	} else if (duration < 120) {
		multiply = 60;
		rate = "/min";
	} else {
		multiply = 3600;
		rate = "/hr";
	}

	snprintf(tmp, sizeof(tmp), "%.2f%s", multiply/duration, rate);
	
	return (tmp);
}

int
file_checkname(char *name)
{
	char *tmp = basename(name);

	return (tmp[0] != '.' && tmp[0] != '/' && !strcmp(tmp, name));
}

/*
 * Copyright (c) 2000 Markus Friedl.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

mysig_t
mysignal(int sig, mysig_t act)
{
#ifdef HAVE_SIGACTION
        struct sigaction sa, osa;

        if (sigaction(sig, NULL, &osa) == -1)
                return (mysig_t) -1;
        if (osa.sa_handler != act) {
                memset(&sa, 0, sizeof(sa));
                sigemptyset(&sa.sa_mask);
                sa.sa_flags = 0;
#if defined(SA_INTERRUPT)
                if (sig == SIGALRM)
                        sa.sa_flags |= SA_INTERRUPT;
#endif
                sa.sa_handler = act;
                if (sigaction(sig, &sa, NULL) == -1)
                        return (mysig_t) -1;
        }
        return (osa.sa_handler);
#else
        return (signal(sig, act));
#endif
}
